local ANIMATION_SPRITE_UPDATE_FREQUENCY = 60
local DEFAULT_FPS = 30

-- Создание класса анимированного спрайта
function CreateAnimationSpriteClass (animation_decls, base_frame_name, fps)
  -- Анимации
  
  local animations = {}
  local base_frame

  -- Загрузка анимации
  local function LoadSprite (frame_name, name, animation)
    local sprite = GetSprite (frame_name)
      
    if not sprite then error (string.format ("Sprite '%s' not found for animation '%s'", frame_name, name)) end
      
    local layout = sprite.layout
      
    if not layout then error (string.format ("Layout '%s' not found for animation '%s'", frame_name, name)) end
      
    local frame = {}
      
    local left     = layout:Get ("Left") + 0
    local top      = layout:Get ("Top") + 0
    local width    = layout:Get ("Width") + 0
    local height   = layout:Get ("Height") + 0
    local opacity  = layout:Get ("Opacity") + 0
    local sprite_x = left + width / 2
    local sprite_y = -(top + height / 2)      
      
    frame.position   = {x = sprite_x, y = sprite_y}
    frame.size       = {x = width, y = height}
    frame.material   = layout:Get ("Material")
    frame.tex_size   = StringToVec (layout:Get ("TexSize"))
    frame.tex_offset = StringToVec (layout:Get ("TexOffset"))
    frame.alpha      = opacity / 255.0
      
    if not base_frame then base_frame = frame end
      
    frame.offset = {x = frame.position.x - base_frame.position.x, y = frame.position.y - base_frame.position.y}
      
    if animation then
      animation.frames [animation.frames_count] = frame
      
      animation.frames_count = animation.frames_count + 1
    end
  end
  
  local function LoadAnimation (name, frames, fps)
    if not name then error ("Name not defined for animation") end
    if not frames then error (string.format ("No frames for animation '%s'", name)) end
  
    if not fps then fps = DEFAULT_FPS end
    
    local animation = {}
    
    animation.frames = {}
    animation.frames_count = 0
    
    for _, frame_name in pairs (frames) do
      LoadSprite (frame_name, animation_name, animation)
    end
    
    animation.fps = fps
    
    animations [name] = animation
  end

  local function SetAnimationScale (name, scale)
    local animation = animations [name]
    
    if not animation then error ("No animation with name '" .. name .. "'") end
    
    if type (scale) == "number" then
      scale = { x = scale, y = scale }
    end
    
    animation.scale = scale
  end

  local function SetAnimationFps (name, fps)
    local animation = animations [name]
    
    if not animation then error ("No animation with name '" .. name .. "'") end
    
    animation.fps = fps
  end

  local function SetAnimationDuration (name, duration)
    local animation = animations [name]
    
    if not animation then error ("No animation with name '" .. name .. "'") end
    
    animation.fps = animation.frames_count / duration
  end
  
  local function GetAnimationDuration (name)
    local animation = animations [name]
    
    if not animation then error ("No animation with name '" .. name .. "'") end
    
    return animation.frames_count / animation.fps
  end
  
  -- Создание анимированного спрайта
  
  local function CreateInstance (parent_node, timer_name)
    local sprite         = Scene.Sprite.Create ()
    local position       = {x = 0, y = 0, z = 0}
    local pivot_position = {x = 0, y = 0}
    local scale          = {x = 1, y = 1}
    local angle          = 0
    local alpha          = 1
    local shown          = false
    local current_animation_name

    if not timer_name then
      timer_name = "default"
    end

    if not parent_node then
      parent_node = game.current_screen.node
    end
    
    sprite.ScalePivotEnabled = false
    
    local current_frame
    
    -- Установка начальных параметров
    
    sprite.Position = vec3 (position.x, position.y, position.z)
    
    -- Изменение положения
    
    local function SetPosition (x, y, z)
      if not z then z = position.z end
    
      if x == position.x and y == position.y and z == position.z then return end
    
      position = {x=x, y=y, z=z}
      
      if current_frame then
        sprite.Position = vec3 (position.x + current_frame.offset.x * scale.x, position.y + current_frame.offset.y * scale.y, position.z)
      end
    end
    
    local function GetPosition ()
      return {x = position.x, y = position.y, z = position.z}
    end
    
    -- Изменение положения точки вращения
    
    local function SetPivotPosition (x, y)
      if x == pivot_position.x and y == pivot_position.y then return end
    
      pivot_position = {x=x, y=y}

      if current_frame then
        sprite.PivotPosition = vec3 (pivot_position.x - current_frame.offset.x / current_frame.size.x, pivot_position.y - current_frame.offset.y / current_frame.size.y, 0)
      end
    end
    
    local function GetPivotPosition ()
      return pivot_position
    end
    
    -- Изменение размера
    
    local function SetScale (sx, sy)
      if sx == scale.x and sy == scale.y then return end
      
      scale = {x=sx, y=sy}
      
      if current_frame then
        sprite.Scale    = vec3 (current_frame.size.x * scale.x, current_frame.size.y * scale.y, 1)
        sprite.Position = vec3 (position.x + current_frame.offset.x * scale.x, position.y + current_frame.offset.y * scale.y, position.z)        
      end
    end
    
    local function GetScale ()
      return {x = scale.x, y = scale.y}
    end
    
    local function GetSize ()
      return {x=current_frame.size.x*scale.x, y=current_frame.size.y*scale.y}
    end

    -- Поворот
    local function SetAngle (new_angle)
      if angle == new_angle then return end
      
      if current_frame then
        angle = new_angle
      
        sprite:SetOrientation (angle, 0, 0, 1)
      end
    end
    
    local function GetAngle ()
      return angle
    end
    
    -- Альфа
    
    local function SetAlpha (in_alpha)
      alpha = in_alpha
      
      if current_frame then
        sprite.Alpha = current_frame.alpha * alpha
      end
    end
    
    local function GetAlpha ()
      return alpha
    end
  
    -- Показ спрайта
    
    local showAction
    
    local function Show (animation_name, on_end_animation, time_offset, pause_duration)
      -- Проверка корректности аргументов

      local animation = animations [animation_name]
      
      if not animation then error (string.format ("Can't find animation '%s'", animation_name)) end
    
      current_animation_name = animation_name
    
      -- Начало проигрывания анимации
      
      if showAction then
        -- Остановка предыдущей анимации
        
        showAction.Stop ()
        
        showAction = nil
      end
      
      local animation_scale = GetTableCopy (animation.scale or { x = 1, y = 1 })
      
      animation_scale.x = scale.x * animation_scale.x
      animation_scale.y = scale.y * animation_scale.y
      
      current_frame = nil      
  
      local current_frame_index = -1
      local is_shown
      local current_material = ""    
      
      -- Старт новой анимации
      
      local function UpdateFrame (frame_index)
        if not is_shown then
          sprite:BindToParent (parent_node, Scene.NodeBindMode.Default, Scene.NodeTransformSpace.Local)
          is_shown = true
        end                
        
        if current_frame_index == frame_index then return end
      
        current_frame_index = frame_index
      
        local frame = animation.frames [frame_index]
        
        current_frame = frame
        
        if not frame then error (string.format ("No frame %u in animation '%s' (frames_count=%u)", frame_index, animation_name, animation.frames_count)) end

        -- Замена параметров спрайта
        
        sprite.Position      = vec3 (position.x + frame.offset.x * animation_scale.x, position.y + frame.offset.y * animation_scale.y, position.z)
        sprite.PivotPosition = vec3 ((pivot_position.x * base_frame.size.x - current_frame.offset.x) / current_frame.size.x, (pivot_position.y * base_frame.size.y - current_frame.offset.y) / current_frame.size.y, 0)
        sprite.Scale         = vec3 (frame.size.x * animation_scale.x, frame.size.y * animation_scale.y, 1)
        sprite.Alpha         = frame.alpha * alpha
        sprite.TexSize       = frame.tex_size
        sprite.TexOffset     = frame.tex_offset

        if frame.material ~= current_material then
          sprite.Material  = frame.material
          current_material = frame.material
        end
      end
      
      local triggered_event_count = 0
      local total_duration        = animation.frames_count / animation.fps + pause_duration
      
      local function Update ()
        -- Получение описания текущего кадра
      
        local time         = showAction.GetTime () + time_offset
        local clamped_time = math.max (0, time % total_duration)
        local frame_index  = math.floor (clamped_time * animation.fps)
        local repeat_index = math.floor (time / total_duration)

        if frame_index >= animation.frames_count then 
          sprite:Unbind ()
          is_shown = false
        else 
          UpdateFrame (frame_index)
        end
        
        if repeat_index > triggered_event_count and on_end_animation then
          triggered_event_count = repeat_index        
          on_end_animation ()          
        end
      end
      
      showAction = StartPeriodicCall (timer_name, 1.0 / ANIMATION_SPRITE_UPDATE_FREQUENCY, Update)
  
      Update ()
    
      -- Помещение спрайта в сцену
      
      if not shown then
        UpdateFrame (0)
        
        shown = true
      end
    end
    
    -- Скрытие спрайта
    
    local function Hide ()
      if not shown then return end
      
      -- Остановка анимации
      
      if showAction then
        showAction.Stop ()
        
        showAction = nil
      end
      
      current_frame = nil      
      
      -- Удаление спрайта из сцены
    
      sprite:Unbind ()
      
      shown = false
    end
    
    -- Работа с очередью анимаций
    
    local animationQueueFirst
    local animationQueueLast
    
    local function OnEndAnimation ()
      -- Определение текущей анимации
    
      local prev_item = animationQueueFirst
      
      if not prev_item then return end
      
      -- Оповещение о конце анимации
      
      if prev_item.on_end_animation then
        prev_item.on_end_animation ()
      end

      -- Определение следующей анимации
      
      if not animationQueueFirst then return end
      
      local item = animationQueueFirst.next 
      
      if not item then item = animationQueueFirst end

      -- Удаление текущей анимации из очереди
      
      if prev_item ~= item then
        animationQueueFirst = item
      
        if not animationQueueFirst then animationQueueLast = nil end
      end
            
      -- Оповещение о начале анимации
      
      if item.on_start_animation then
        item.on_start_animation ()
      end
      
      Show (item.animation_name, OnEndAnimation, item.time_offset, item.pause_duration)
      
      item.time_offset = 0
    end    
    
    local function Enqueue (animation_name, on_start_animation, on_end_animation, time_offset, pause_duration)
      if not animations [animation_name] then error (string.format ("Animation '%s' not found", animation_name)) end
    
      local item = {}
      
      item.animation_name     = animation_name
      item.on_start_animation = on_start_animation
      item.on_end_animation   = on_end_animation
      item.time_offset        = time_offset or 0
      item.pause_duration     = pause_duration or 0
      
      if animationQueueLast then animationQueueLast.next = item
      else                       animationQueueFirst     = item end
      
      animationQueueLast = item
      
      if not shown then
        Show (item.animation_name, OnEndAnimation, item.time_offset, item.pause_duration)
        
        item.time_offset = 0
        
        if (on_start_animation) then
          on_start_animation ()
        end
      end
    end    
    
    local function ResetQueue ()
      animationQueueFirst = nil
      animationQueueLast  = nil
    end
    
    local function GetNode ()
      return sprite
    end

    local function SetParent (parent)
      parent_node = parent
      
      if sprite.Parent then
        sprite:BindToParent (parent)
      end
    end
      
    -- Методы экземпляра анимированного спрайта видимые снаружи   
  
    local methods = {}
    
    methods.Show = function (animation_name, on_end_animation, time_offset, pause_duration)
      ResetQueue ()
      Hide ()
      Enqueue (animation_name, nil, on_end_animation, time_offset, pause_duration)
    end
    
    methods.Hide = function ()
      ResetQueue ()
      Hide ()
    end
    
    methods.Enqueue                 = Enqueue    
    methods.SetPosition             = SetPosition
    methods.GetPosition             = GetPosition
    methods.SetPivotPosition        = SetPivotPosition
    methods.GetPivotPosition        = GetPivotPosition
    methods.SetScale                = SetScale    
    methods.GetScale                = GetScale
    methods.SetAlpha                = SetAlpha
    methods.GetAlpha                = GetAlpha
    methods.GetSize                 = GetSize
    methods.SetAngle                = SetAngle
    methods.GetAngle                = GetAngle
    methods.SetAnimationFps         = SetAnimationFps 
    methods.GetAnimationDuration    = GetAnimationDuration
    methods.SetAnimationDuration    = SetAnimationDuration 
    methods.SetAnimationScale       = SetAnimationScale 
    methods.GetNode                 = GetNode
    methods.SetParent               = SetParent
    methods.GetCurrentAnimationName = function () return current_animation_name end

    return methods    
  end
  
  -- Инициализация

  if base_frame_name then
    LoadSprite (base_frame_name, "")
  end
  
  if animation_decls then
    for name, value in pairs (animation_decls) do
      LoadAnimation (name, value, fps)  
    end
  end  

  -- Методы класса, видимые снаружи
  
  local methods = {}
  
  methods.LoadAnimation        = LoadAnimation
  methods.CreateInstance       = CreateInstance
  
  return methods
end
