HS_DEV_PORT  = "Dev"
HS_QA_PORT   = "QA"
HS_LIVE_PORT = "Live"

local HS_PORT            = HS_LIVE_PORT
local MAX_FRIENDS_COUNT  = 400
local MOCKUP_ENABLED     = false
local RECONNECT_DELAY    = 0.5
local DISCONNECT_ON_DEACTIVATE_DELAY = 10 * 60

-- Stored messages IDs
AUTH_MESSAGE        = "user.authenticate"
BONUS_INIT_MESSAGE  = "bonus.init"
BONUS_MONEY_MESSAGE = "bonus.money"
SLOT_INIT_MESSAGE   = "slot.init"

SLOT_FEATURE_DOUBLE_GAME       = 1
SLOT_FEATURE_DOUBLE_WILD       = 2
SLOT_FEATURE_TRIPLE_WILD       = 4
SLOT_FEATURE_BONUS_ANYWHERE    = 8
SLOT_FEATURE_EXPANDED_WILD     = 16
SLOT_FEATURE_FREE_SPINS_SELECT = 32
SLOT_FEATURE_BONUS_SPINS       = 64

local ERROR_CODE_BONUS_NOT_ACTIVE               = 109
local ERROR_CODE_SLOT_LEVEL_TOO_HIGH            = 115
local ERROR_CODE_SLOT_FOR_BONUS_NOT_INITIALIZED = 138

local DUMMY_MESSAGE = "lobby.init"

local connected         = false
local connecting        = false
local connect_initiated = false

local waiting_for_slot_init_response

local ignored_errors   = {}
local stored_messages  = {}
local stored_responses = {}

local user_id
local user_diamonds
local user_money
local user_wagering
local battles
local battles_timeout
local bets
local current_slot_id
local friends
local grant_level
local grant_time
local grant_period
local ignore_connection_errors
local ignore_next_balance_update
local install_bonus
local is_depositor
local is_platform_depositor
local lobby_bonuses
local login_server
local migration_details
local needs_show_open_doors_promo_ended_action
local needs_show_promo_ended_action
local open_doors_promo_ended_action
local payments
local payments_data_node
local payments_diamonds
local payments_diamonds_data_node
local payments_level_bonuses
local powerups_store
local powerups
local prizes
local promo_actions
local promo_ended_action
local slots
local user_levels
local user_level
local top_daily_players
local tutorials
local sent_gifts
local welcome_bonus
local wheel_data

local was_connected = false

local disconnect_on_deactivate
local reconnecting
local authenticating
local reconnect_window_shown
local tried_reconnect_after_error

local is_active
local active_battle_search_id

function GetHsPort ()
  return HS_PORT
end

function SetHsPort (port)
  HS_PORT = port
end

local function ConnectionStateName (state)
  if state == 0 then
    return "Disconnected"
  elseif state == 1 then
    return "Connecting"
  elseif state == 2 then
    return "Connected" 
  else
    return "Unknown"
  end
end

local function ErrorCodeName (code)
  if code == 0 then
    return "Generic"
  elseif code == 1 then
    return "SendQueueOverflow" 
  elseif code == 2 then
    return "Timeout" 
  elseif code == 3 then
    return "InvalidHeader" 
  elseif code == 4 then
    return "HashMismatch" 
  else
    return "Unknown"
  end
end

local function ClearHsConnectionData ()
  waiting_for_slot_init_response = nil
  ignored_errors = {}

  user_id                     = 0
  user_diamonds               = -1
  user_money                  = -1
  user_wagering               = -1
  battles                     = {}
  battles_timeout             = 0
  bets                        = {}
  current_slot_id             = nil
  migration_details           = {}
  slots                       = {}
  payments                    = {}
  payments_diamonds           = {}
  payments_diamonds_data_node = nil
  payments_level_bonuses      = {}
  payments_data_node          = nil
  prizes                      = {}
  promo_actions               = {}
  user_levels                 = {}
  user_level                  = nil
  grant_level                 = 0
  grant_time                  = nil
  grant_period                = 0
  ignore_next_balance_update  = nil
  install_bonus               = nil
  is_depositor                = nil
  is_platform_depositor       = nil
  lobby_bonuses               = {}
  login_server                = nil
  top_daily_players           = {}
  tutorials                   = {}
  friends                     = {}
  sent_gifts                  = {}
  powerups_store              = {}
  powerups                    = {}
  welcome_bonus               = nil
  wheel_data                  = nil

  ApplicationDataSet (BATTLE_INIT_DATA, nil)
  ApplicationDataSet (NEEDS_SHOW_NO_MONEY_WINDOW_ON_LOBBY_RETURN, nil)
  ApplicationDataSet (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS, nil)
  ApplicationDataSet (LOBBY_SHOW_FREE_SPINS_TAKE_GIFT_ID, nil)
end

function OnHsConnectionStateChanged (state)
  print ("Connection state changed " .. ConnectionStateName (state))
  
  connecting = state == 1
  connected  = state == 2
  
  if not was_connected then
    was_connected = was_connected or connected
  end
  
  local background_mode = System.ApplicationBackgroundMode.Suspend
  
  if connected then
    background_mode = System.ApplicationBackgroundMode.Active

    tried_reconnect_after_error = false

    ApplicationDataSet (PROMO_ACTIONS_SHOWN, false)
  else
    ApplicationDataSet (LAST_GAMES_RESOURCES_DOWNLOAD_TIME, nil) --force reload index on reconnect
  end

  System.Application.BackgroundMode = background_mode
  
  if connected and reconnect_window_shown then
    game.scripts.ui_windows.HideWindow ()
  end
end

function HsConnectionIsConnected ()
  return MOCKUP_ENABLED or connected
end

local function ParseUserLevel (level_node, is_current)
  local level = {}
  
  level.id          = level_node:Find ("id"):Attribute (0) + 0
  level.wagering    = level_node:Find ("wagering"):Attribute (0) + 0
  level.max_bet_id  = level_node:Find ("maxBetId"):Attribute (0) + 0

  local function ParseBonusValue (bonus_node)
    return { coins = GetNodeChildAttribute (bonus_node, "coins", 0) + 0, diamonds = GetNodeChildAttribute (bonus_node, "diamonds", 0) + 0 }
  end

  level.bonus       = ParseBonusValue (level_node:Find ("bonus"):Child (0))
  level.mega_bonus  = ParseBonusValue (level_node:Find ("megaBonus"):Child (0))
  level.super_bonus = ParseBonusValue (level_node:Find ("superBonus"):Child (0))
  
  user_levels [level.id] = level
  
  if is_current then
    user_level = level
  end
end

local function ParsePerson (node)
  local person = {}
  
  person.first_name      = GetNodeChildAttribute (node, "firstName", "")
  person.last_name       = GetNodeChildAttribute (node, "lastName", "")
  person.photo           = GetNodeChildAttribute (node, "photo", "")
  person.money           = GetNodeChildAttribute (node, "money", 0) + 0
  person.level           = GetNodeChildAttribute (node, "level", 0) + 0
  person.user_id         = GetNodeChildAttribute (node, "userId", 0) + 0
  person.login_id        = GetNodeChildAttribute (node, "loginId", "")
  person.achiev_group_id = GetNodeChildAttribute (node, "achievGroupId", 0) + 0
  
  if user_id and person.user_id == user_id then
    person.level = user_level.id + 0
  end
  
  if person.achiev_group_id > ACHIEVEMENTS_GROUPS_COUNT then
    person.achiev_group_id = 0
  end
  
  return person
end

function HsConnectionIsAllStoredMessagesLoaded ()
  for i = 1, #stored_messages do
    if stored_responses [stored_messages [i].message_id] == nil then
      print ("message " .. stored_messages [i].message_id .. " not loaded")
    
      return false 
    end
  end
  
  return true
end

local function HsConnectionOnStoredMessageLoaded (message_id, event)
  stored_responses [message_id] = event

  if HsConnectionIsAllStoredMessagesLoaded () then
    if reconnecting then
      reconnecting = false
    
      local last_message = stored_messages [#stored_messages]

      if last_message.reconnect_handler then
        last_message.reconnect_handler ()
      end
    end
  
    ExecuteOnNextUpdate (function ()
      if game.scripts.level and game.scripts.level.OnAllStoredMessagesReceived then
        game.scripts.level.OnAllStoredMessagesReceived ()
      end
    end)
  end
end

local function ShowUpdateVersionAlert ()
  if not ApplicationDataGet (LATEST_PROTOCOL_ENGINE_VERSION) then
    DelayedAction ("ui", 1, ShowUpdateVersionAlert)
    return
  end
  
  if PROTOCOL_ENGINE_VERSION < ApplicationDataGet (LATEST_PROTOCOL_ENGINE_VERSION) then
    if game.scripts.ui_windows then
      DelayedAction ("ui", 1, function ()
        game.scripts.ui_windows.ShowUpdateApplicationWindow ()
      end)
    end
  end
    
  if game.scripts.ui_windows then
    DelayedAction ("ui", 1, function ()
      LogFlurryEvent ("ReconnectPopupShown", { Reason = "Error 101" }, true)  
      game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
        LogFlurryEvent ("ReconnectPopupClick", { Reason = "Error 101" }, true)  
        HsConnectionReconnect ()
      end, nil, GetLocalizedString ("RECONNECT"), { GetLocalizedString ("CONNECTION_ERROR"), "101" })
    end)
  end
end

local function ClearStoredResponse (message_id)
  stored_responses [message_id] = nil
  
  if message_id == BONUS_INIT_MESSAGE then
    stored_responses [BONUS_MONEY_MESSAGE] = nil
  end
end

local HS_MESSAGE_PARSE_FUNCTIONS = {}

HS_MESSAGE_PARSE_FUNCTIONS ["achievement.complete"] = function (event)
  local achievement_data = {}
  
  achievement_data.id         = GetNodeChildAttribute (event, "id") + 0
  achievement_data.group_id   = GetNodeChildAttribute (event, "groupId") + 0
  achievement_data.current    = GetNodeChildAttribute (event, "current", 0) + 0
  achievement_data.prize_type = GetNodeChildAttribute (event, "prizeType", 0) + 0

  local achievement_event_processed

  if game.scripts.level and game.scripts.level.OnAchievementComplete then
    achievement_event_processed = game.scripts.level.OnAchievementComplete (achievement_data)
  end 
  
  if game.scripts.lobby then
    game.scripts.lobby.OnAchievementComplete (achievement_data)
    achievement_event_processed = true
  end 
  
  if not achievement_event_processed then
    local pending_achievement_events = ApplicationDataGet (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS) or {}
    
    table.insert (pending_achievement_events, { event = "achievement.complete", data = achievement_data })
    
    ApplicationDataSet (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS, pending_achievement_events)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.cancelinvite"] = function (event)
  if game.scripts.ui_windows then
    game.scripts.ui_windows.OnBattleInviteCanceled (GetNodeChildAttribute (event, "userId", 0) + 0)
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.info"] = function (event)
  if game.scripts.level and game.scripts.level.OnBattleInfoUpdated then
    game.scripts.level.OnBattleInfoUpdated (event)
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.init"] = function (event)
  local battles_node  = event:Find ("battles"):Child (0)
  local battles_count = battles_node.ChildrenCount

  battles = {}
  battles_timeout = GetNodeChildAttribute (event, "timeout", 0) + 0

  for i = 0, battles_count - 1 do
    local battle_node = battles_node:Child (i)
    
    local battle = {}
    
    battle.slot_id   = GetNodeChildAttribute (battle_node, "slotId") + 0
    battle.data_node = battle_node
    battle.modes     = {}

    local modes_node  = battle_node:Find ("modes"):Child (0)
    local modes_count = math.min (3, modes_node.ChildrenCount)
    
    if modes_count < 3 then
      print ("Battle with modes count " .. modes_count .. " ignored")
    else
      for j = 0, modes_count - 1 do
        local mode_node = modes_node:Child (j)
        
        local mode = {}
        
        mode.id             = GetNodeChildAttribute (mode_node, "id") + 0
        mode.time           = GetNodeChildAttribute (mode_node, "time", 0) + 0
        mode.bet_id         = GetNodeChildAttribute (mode_node, "betId") + 0
        mode.coins_fee      = GetNodeChildAttribute (mode_node, "coinsFee", 0) + 0
        mode.level          = GetNodeChildAttribute (mode_node, "level") + 0
        mode.win_type       = GetNodeChildAttribute (mode_node, "winType") + 0
        mode.win_money      = GetNodeChildAttribute (mode_node, "winMoney", 0) + 0
        
        table.insert (battle.modes, mode)
      end
    
      table.insert (battles, battle)
    end
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.interrupt"] = function (event)
  if waiting_for_slot_init_response then
    ignored_errors ["145"] = true --ignore error about slot.init after battle.interrupt
    
    current_slot_id = nil
    
    for i = 1, #stored_messages do
      if stored_messages [i].message_id == SLOT_INIT_MESSAGE then
        ClearStoredResponse (SLOT_INIT_MESSAGE)
        stored_messages [i] = nil
        break
      end
    end
  end

  ApplicationDataSet (BATTLE_INTERRUPTED, true)

  if game.scripts.level and game.scripts.level.OnBattleInterrupt then
    game.scripts.level.OnBattleInterrupt ()
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.invite"] = function (event)
  if game.current_screen.name ~= "lobby" and not ApplicationDataGet (IS_BATTLE_ACTIVE) then return end --invites ignored in single player mode

  local invite_data = {}
  
  invite_data.mode_id  = GetNodeChildAttribute (event, "modeId") + 0
  invite_data.opponent = {}
  
  local opponent_node = event:Find ("opponent"):Child (0)
  
  invite_data.opponent.user_id    = GetNodeChildAttribute (opponent_node, "userId") + 0
  invite_data.opponent.level      = GetNodeChildAttribute (opponent_node, "level", 1) + 0
  invite_data.opponent.first_name = GetNodeChildAttribute (opponent_node, "firstName", "")
  invite_data.opponent.last_name  = GetNodeChildAttribute (opponent_node, "lastName", "")
  invite_data.opponent.photo      = GetNodeChildAttribute (opponent_node, "photo", "")
  invite_data.opponent.wins       = GetNodeChildAttribute (opponent_node, "wins", 0) + 0
  invite_data.opponent.loses      = GetNodeChildAttribute (opponent_node, "loses", 0) + 0
  invite_data.opponent.rating     = GetNodeChildAttribute (opponent_node, "rating", 0) + 0  

  if game.scripts.level and game.scripts.level.OnBattleInvite then
    game.scripts.level.OnBattleInvite (invite_data)
  elseif game.scripts.lobby then
    game.scripts.lobby.OnBattleInvite (invite_data)
  elseif game.current_screen.name == "loader" then
    ApplicationDataSet (BATTLE_INVITE_DATA, invite_data)
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.loading"] = function (event)
  local progress = GetNodeChildAttribute (event, "progress", 0) / 100
  
  ApplicationDataSet (BATTLE_OPPONENT_LOADING_PROGRESS, progress)
  
  if game.scripts.level and game.scripts.level.SetOpponentLoadingProgress then
    game.scripts.level.SetOpponentLoadingProgress (progress)
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.result"] = function (event)
  local status      = GetNodeChildAttribute (event, "status", 0) + 0
  local money       = GetNodeChildAttribute (event, "money", 0) + 0
  local rating      = GetNodeChildAttribute (event, "rating", 0) + 0
  local interrupted = GetNodeChildAttribute (event, "interrupted", "false") == "true"

  if game.scripts.level and game.scripts.level.OnBattleResult then
    game.scripts.level.OnBattleResult (status, money, rating, interrupted)
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.search"] = function (event)
  local search_id = GetNodeChildAttribute (event, "guid", 0) + 0
  
  if search_id == active_battle_search_id then
    local status = GetNodeChildAttribute (event, "status", 0) + 0
    
    if status == 1 and game.scripts.ui_windows then
      local function ParsePlayerInfo (node)
        if node and node.ChildrenCount > 0 then
          node = node:Child (0)
        end
      
        local info = {}
        
        info.user_id    = GetNodeChildAttribute (node, "userId") + 0
        info.first_name = GetNodeChildAttribute (node, "firstName", "")
        info.last_name  = GetNodeChildAttribute (node, "lastName", "")
        info.photo      = GetNodeChildAttribute (node, "photo", "")
        info.level      = GetNodeChildAttribute (node, "level", 1) + 0
        info.wins       = GetNodeChildAttribute (node, "wins", 0) + 0
        info.loses      = GetNodeChildAttribute (node, "loses", 0) + 0
        
        if info.photo ~= "" and string.find (info.photo, "http") ~= 1 then
          info.photo = login_server .. info.photo
        end
        
        return info
      end
    
      local battle_info = { you = ParsePlayerInfo (event:Find ("you")), opponent = ParsePlayerInfo (event:Find ("opponent")) } 

      ApplicationDataSet (BATTLE_PLAYERS_INFO, battle_info)
    
      game.scripts.ui_windows.OnBattleSearchSucceeded (event)
    elseif status == 2 and game.scripts.ui_windows then
      game.scripts.ui_windows.HideWindow ()
      game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, 
        { GetLocalizedString ("BATTLES_LOCKED_ERROR_LINE1"), GetLocalizedString ("BATTLES_LOCKED_ERROR_LINE2"),
          GetLocalizedString ("BATTLES_LOCKED_ERROR_LINE3"), GetLocalizedString ("BATTLES_LOCKED_ERROR_LINE4"),
          GetLocalizedString ("BATTLES_LOCKED_ERROR_LINE5") })
    end
  end
  
  active_battle_search_id = nil
end

HS_MESSAGE_PARSE_FUNCTIONS ["battle.start"] = function (event)
  if game.scripts.level and game.scripts.level.OnBattleStart then
    game.scripts.level.OnBattleStart ()
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["bonus.exit"] = function (event)
  print ('-------bonus.exit--------')
end

HS_MESSAGE_PARSE_FUNCTIONS ["bonus.init"] = function (event)
  --Do nothing
end

HS_MESSAGE_PARSE_FUNCTIONS ["bonus.money"] = function (event)
  ApplicationDataSet (BONUS_GAME_MONEY, event:Find ('money'):Attribute(0))

  if game.scripts.level and game.scripts.level.OnBonusMoneyResponse then
    game.scripts.level.OnBonusMoneyResponse (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["bonus.pick"] = function (event)
  if needs_show_promo_ended_action then
    game.scripts.ui_windows.ShowHappyTimeActionEndedWindow (needs_show_promo_ended_action, function ()
      SetCurrentScreen ("lobby")
    end)
  end

  if needs_show_open_doors_promo_ended_action then
    game.scripts.ui_windows.ShowOpenDoorsActionIsOver (needs_show_open_doors_promo_ended_action)
  end

  if game.scripts.level and game.scripts.level.OnBonusPickResponse then
    game.scripts.level.OnBonusPickResponse (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["diamonds.update"] = function (event)
  payments_diamonds = {}
  payments_diamonds_data_node = event

  local diamonds_node  = event:Find ("payments"):Child (0)
  local diamonds_count = diamonds_node.ChildrenCount

  for i = 0, diamonds_count - 1 do
    local diamonds_node = diamonds_node:Child (i)

    local diamond = {}

    diamond.money             = diamonds_node:Find ("money"):Attribute (0) + 0
    diamond.bonus             = diamonds_node:Find ("bonus"):Attribute (0) + 0
    diamond.total_money       = diamonds_node:Find ("totalMoney"):Attribute (0) + 0
    diamond.total_bonus_money = diamonds_node:Find ("totalBonusMoney"):Attribute (0) + 0
    diamond.product_id        = diamonds_node:Find ("productId"):Attribute (0)
    diamond.network_money     = diamonds_node:Find ("networkMoney"):Attribute (0) + 0
    diamond.data_node         = diamonds_node

    table.insert (payments_diamonds, diamond)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["double.exit"] = function (event)
  if game.scripts.level and game.scripts.level.OnDoubleExitResponse then
    game.scripts.level.OnDoubleExitResponse (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["double.init"] = function (event)
  if game.scripts.level and game.scripts.level.OnDoubleInitResponse then
    game.scripts.level.OnDoubleInitResponse (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["double.try"] = function (event)
  if game.scripts.level and game.scripts.level.OnDoubleTryResponse then
    game.scripts.level.OnDoubleTryResponse (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["gifts.sent"] = function (event)
  local sent_gifts_node = event:Find ("sentGifts"):Child (0)

  sent_gifts = {}

  for i = 0, sent_gifts_node.AttributesCount - 1 do
    table.insert (sent_gifts, sent_gifts_node:Attribute (i))
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["gifts.update"] = function (event)
  local sent_gifts_node = event:Find ("sentGifts")

  if sent_gifts_node then
    sent_gifts_node = sent_gifts_node:Child (0)

    sent_gifts = {}

    for i = 0, sent_gifts_node.AttributesCount - 1 do
      table.insert (sent_gifts, sent_gifts_node:Attribute (i))
    end
  end

  ApplicationDataSet (GIFTS_EVENT, event)

  if game.scripts.lobby then
    game.scripts.lobby.OnGiftsUpdate (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["grant.mega"] = function (event)
  if game.scripts.lobby then
    game.scripts.lobby.OnMegaGrantTaken (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["grant.simple"] = function (event)
  if game.scripts.lobby then
    game.scripts.lobby.OnSimpleGrantTaken (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["grant.update"] = function (event)
  local time_to_grant = event:Find ("time"):Attribute (0) + 0

  grant_level = event:Find ("level"):Attribute (0) + 0
  grant_time  = time_to_grant + game.main_timer.GetTime ()

  if game.scripts.lobby then
    game.scripts.lobby.OnGrantUpdate ()
  end
  
  if time_to_grant > 0 then
    System.Application.PostNotification ("LocalNotifications Schedule|Bonus|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.BONUS_ACTIVE") .. "|" .. time_to_grant .. "|" .. time_to_grant + 24 * 60 * 60 .. "|" .. time_to_grant + 24 * 60 * 60 * 3 .. "|" .. time_to_grant + 24 * 60 * 60 * 7)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["friends.update"] = function (event)
  local friends_node  = event:Find ("friends"):Child (0)
  local friends_count = friends_node.ChildrenCount
  
  friends = {}

  for j = 0, math.min (MAX_FRIENDS_COUNT, friends_count - 1) do
    table.insert (friends, ParsePerson (friends_node:Child (j)))
  end

  --TODO update something        
end

HS_MESSAGE_PARSE_FUNCTIONS ["info.achievements"] = function (event)
  local achievement_data = {}
  
  achievement_data.id         = GetNodeChildAttribute (event, "id") + 0
  achievement_data.prize      = GetNodeChildAttribute (event, "prize", 0) + 0
  achievement_data.prize_type = GetNodeChildAttribute (event, "prizeType", 0) + 0

  local achievement_event_processed

  if game.scripts.level and game.scripts.level.OnAchievementGroupComplete then
    achievement_event_processed = game.scripts.level.OnAchievementGroupComplete (achievement_data)
  end 

  if game.scripts.lobby then
    game.scripts.lobby.OnAchievementGroupComplete (achievement_data)
    
    achievement_event_processed = true
  end 
  
  if not achievement_event_processed then
    local pending_achievement_events = ApplicationDataGet (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS) or {}
    
    table.insert (pending_achievement_events, { event = "info.achievements", data = achievement_data })
    
    ApplicationDataSet (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS, pending_achievement_events)
  end
  
  ignore_next_balance_update = true
end

HS_MESSAGE_PARSE_FUNCTIONS ["lobby.init"] = function (event)
  local payment_url = event:Find ("loginServerUrl"):Attribute (0)
  
  if payment_url then
    local base_url_end

    if #payment_url > 9 then
      base_url_end = string.find (payment_url, "/", 9, true)
    end    
    
    if base_url_end then
      login_server = string.sub (payment_url, 1, base_url_end - 1)
    else
      login_server = payment_url
    end
  end
  
  GetStoreManager ().OnPaymentUrlUpdated (payment_url)

  grant_period = event:Find ("grantTime"):Attribute (0) + 0 

  local bets_node  = event:Find ("bets"):Child (0)
  local bets_count = bets_node.ChildrenCount

  for j = 0, bets_count - 1 do
    local bet_node = bets_node:Child (j)
    
    local bet = {}
    
    bet.id    = bet_node:Find ("id"):Attribute (0) + 0
    bet.money = bet_node:Find ("money"):Attribute (0) + 0
    
    table.insert (bets, bet)
  end
  
  local bonuses_node = event:Find ("bonus")
  
  if bonuses_node and bonuses_node.ChildrenCount > 0 then
    bonuses_node = bonuses_node:Child (0)
  
    local mobile_version_install_bonus = bonuses_node:Find ("installMobileVersionBonus")
  
    if mobile_version_install_bonus then
      install_bonus = mobile_version_install_bonus:Attribute (0) + 0
    end
    
    local welcome_coins_bonus    = bonuses_node:Find ("welcomeCoins")
    local welcome_diamonds_bonus = bonuses_node:Find ("welcomeDiamonds")

    if welcome_coins_bonus or welcome_diamonds_bonus then
      welcome_bonus = { coins = 0, diamonds = 0 }
    end
    
    if welcome_coins_bonus then
      welcome_bonus.coins = welcome_coins_bonus:Attribute (0) + 0
    end
    
    if welcome_diamonds_bonus then
      welcome_bonus.diamonds = welcome_diamonds_bonus:Attribute (0) + 0
    end
  end
  
  lobby_bonuses = {}
  
  local bonuses_node = event:Find ("bonuses")
  
  if bonuses_node and bonuses_node.ChildrenCount > 0 then
    local bonuses_node  = bonuses_node:Child (0)
    local bonuses_count = bonuses_node.ChildrenCount
    
    for i = 0, bonuses_count - 1 do
      local bonus_node = bonuses_node:Child (i)
      
      local bonus = {}
      
      bonus.type     = GetNodeChildAttribute (bonus_node, "type") + 0
      bonus.coins    = GetNodeChildAttribute (bonus_node, "coins", 0) + 0
      bonus.diamonds = GetNodeChildAttribute (bonus_node, "diamonds", 0) + 0
      bonus.expired  = GetNodeChildAttribute (bonus_node, "expired", "false") == "true"
      
      local bonus_blocked = false
      
      if bonus.type == 3 and not ApplicationDataGet (GOOGLE_PLAY_SERVICES_STATUS) then
        bonus_blocked = true
      end
      
      if not bonus_blocked then
        table.insert (lobby_bonuses, bonus)
      end
    end
  end 
  
  tutorials = GetNodeChildArray (event, "tutorials", {})
end

HS_MESSAGE_PARSE_FUNCTIONS ["migration.complete"] = function (event)
  SetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE)
end

HS_MESSAGE_PARSE_FUNCTIONS ["migration.details"] = function (event)
  migration_details.money    = GetNodeChildAttribute (event, "money", 0) + 0
  migration_details.level    = GetNodeChildAttribute (event, "level", 0) + 0
  migration_details.diamonds = GetNodeChildAttribute (event, "diamonds", 0) + 0
  migration_details.merged   = GetNodeChildAttribute (event, "merged", "false") == "true"
  migration_details.success  = GetNodeChildAttribute (event, "success", "false") == "true"
  
  if not migration_details.success then
    SetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["payment.check"] = function (event)
  is_depositor = true

  DelayedAction ("ui", 0.1, function ()
    if game.scripts.top_ui then
      game.scripts.top_ui.MoneyForceUpdate (user_money)
    end
  end)
end

HS_MESSAGE_PARSE_FUNCTIONS ["payments.update"] = function (event)
  payments_data_node     = event
  payments               = {}
  payments_level_bonuses = {}
  
  local payments_node  = event:Find ("payments"):Child (0)
  local payments_count = payments_node.ChildrenCount
  
  for i = 0, payments_count - 1 do
    local payment_node = payments_node:Child (i)
    
    local payment = {}

    payment.money             = payment_node:Find ("money"):Attribute (0) + 0
    payment.bonus             = payment_node:Find ("bonus"):Attribute (0) + 0
    payment.total_money       = payment_node:Find ("totalMoney"):Attribute (0) + 0
    payment.total_bonus_money = payment_node:Find ("totalBonusMoney"):Attribute (0) + 0
    payment.product_id        = payment_node:Find ("productId"):Attribute (0)
    payment.network_money     = payment_node:Find ("networkMoney"):Attribute (0) + 0
    payment.data_node         = payment_node
    
    table.insert (payments, payment)
  end
          
  local payments_bonuses_node  = event:Find ("bonuses"):Child (0)
  local payments_bonuses_count = payments_bonuses_node.ChildrenCount
  
  for i = 0, payments_bonuses_count - 1 do
    local payment_bonus_node = payments_bonuses_node:Child (i)
    
    local bonus = {}

    bonus.level_from = payment_bonus_node:Find ("levelFrom"):Attribute (0) + 0
    bonus.bonus      = payment_bonus_node:Find ("bonus"):Attribute (0) + 0
    
    table.insert (payments_level_bonuses, bonus)
  end        
end

HS_MESSAGE_PARSE_FUNCTIONS ["powerup.store"] = function (event)
  powerups_store = {}
 
  local powerups_node  = event:Find ("powerups"):Child (0)
  local powerups_count = powerups_node.ChildrenCount

  for i = 0, powerups_count - 1 do
    local powerup_node = powerups_node:Child (i)
    
    local powerup = {}
    
    powerup.id       = powerup_node:Find ("id"):Attribute (0) + 0
    powerup.type     = powerup_node:Find ("type"):Attribute (0) + 0
    powerup.price    = powerup_node:Find ("price"):Attribute (0) + 0
    powerup.value    = GetNodeChildAttribute (powerup_node, "value", 1) + 0
    powerup.duration = GetNodeChildAttribute (powerup_node, "duration", 1) + 0

    table.insert (powerups_store, powerup)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["powerup.update"] = function (event)
  powerups = powerups or {}
 
  local powerups_node  = event:Find ("powerups"):Child (0)
  local powerups_count = powerups_node.ChildrenCount

  for i = 0, powerups_count - 1 do
    local powerup_node = powerups_node:Child (i)
    
    local powerup = {}
    
    powerup.id         = powerup_node:Find ("id"):Attribute (0) + 0
    powerup.count      = GetNodeChildAttribute (powerup_node, "count", 0) + 0
    powerup.enable     = GetNodeChildAttribute (powerup_node, "enable", "false") == "true"
    powerup.spins_left = GetNodeChildAttribute (powerup_node, "spinsLeft", 0) + 0

    local powerup_found

    for j = 1, #powerups do
      local current_powerup = powerups [j]
      
      if current_powerup.id == powerup.id then
        powerups [j] = powerup
        
        powerup_found = true
        
        break
      end
    end

    if not powerup_found then
      table.insert (powerups, powerup)
    end
  end
  
  if game.scripts.lobby then
    game.scripts.lobby.OnPowerupUpdated ()
  end
  
  if game.scripts.boosters_overlay then
    game.scripts.boosters_overlay.OnPowerupUpdated ()
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["prize.update"] = function (event)
  prizes = {}

  local prizes_node = event:Find ("prizes")
  
  if not prizes_node or prizes_node.ChildrenCount == 0 then return end

  prizes_node = prizes_node:Child (0)
  
  local prizes_count = prizes_node.ChildrenCount
  
  for i = 0, prizes_count - 1 do
    local prize_node = prizes_node:Child (i)
    local prize      = {}
    
    prize.id       = prize_node:Find ("id"):Attribute (0) + 0
    prize.type     = prize_node:Find ("type"):Attribute (0) + 0
    prize.place    = prize_node:Find ("place"):Attribute (0) + 0
    prize.coins    = GetNodeChildAttribute (prize_node, "coins", 0) + 0
    prize.diamonds = GetNodeChildAttribute (prize_node, "diamonds", 0) + 0
    
    if prize.type == 0 or prize.type == 1 or prize.type == 2 then --ignore new prize types
      table.insert (prizes, prize)
    end
  end

  ShowPendingPrizes ()  
end

local function AddPromoActions (event, add_event)
  local actions_node  = event:Find ("promoActions"):Child (0)
  local actions_count = actions_node.ChildrenCount

  if add_event then
    ApplicationDataSet (PROMO_ACTIONS_SHOWN, false)
  end
  
  local current_time = game.main_timer.GetTime ()
  
  for i = 0, actions_count - 1 do
    local action_node = actions_node:Child (i)
    
    local action = {}

    action.unique_id    = action_node:Find ("id"):Attribute (0) + 0
    action.id           = action_node:Find ("promoId"):Attribute (0) + 0
    action.type         = action_node:Find ("type"):Attribute (0) + 0
    action.time_left    = action_node:Find ("time"):Attribute (0) + 0
    action.time_end     = current_time + action.time_left 
    action.additional   = action_node:Find ("additional")
    action.data_node    = action_node

    if action_node:Find ("priority") then
      action.priority  = action_node:Find ("priority"):Attribute (0) + 0
    else
      action.priority = #promo_actions
    end

    table.insert (promo_actions, action)
    
    if add_event then
      if action.id == 28 then  --new happy time promo
        local status, err = pcall (function ()
          if game.current_screen.name == "loader" or ApplicationDataGet (IS_BATTLE_ACTIVE) then
            ApplicationDataSet (PENDING_HAPPY_HOUR_ACTION, action)
          else
            local additional = action.additional:Child (0)
            local type       = GetNodeChildAttribute (additional, "type") + 0

            if current_slot_id and GetNodeChildAttribute (additional, "value1") + 0 == current_slot_id then
            --ignore this promo
            else
              if game.scripts.lobby and not game.scripts.lobby.IsPromoOpened () and not game.scripts.ui_windows.IsAlertWindowShown () then
                game.scripts.lobby.OpenPromoActions (true)
              elseif not game.scripts.lobby or not game.scripts.lobby.IsPromoOpened () then
                game.scripts.ui_windows.ShowHappyHourWindowAction (action)
              end
            end
          end
        end)

        if not status then
          print (err)
        end
      else --new other promo
        ExecuteOnNextUpdate (function ()
          if game.scripts.lobby and not game.scripts.lobby.IsPromoOpened () and not game.scripts.ui_windows.IsAlertWindowShown () then
            game.scripts.lobby.OpenPromoActions (true)
          end
        end)
      end
    end
  end
  
  table.sort (promo_actions, function (action_1, action_2)
    return action_1.priority < action_2.priority
  end)
  
  if game.scripts.lobby then
    game.scripts.lobby.OnPromoActionsUpdated ()
  end
  
  if game.scripts.buy_chips_window and game.scripts.buy_chips_window.OnPromoActionsUpdated then
    game.scripts.buy_chips_window.OnPromoActionsUpdated ()
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["promo.add"] = function (event)
  AddPromoActions (event, true)  
end

HS_MESSAGE_PARSE_FUNCTIONS ["promo.init"] = function (event)
  promo_actions = {}

  AddPromoActions (event)  
end

HS_MESSAGE_PARSE_FUNCTIONS ["promo.remove"] = function (event)
  local ids_node  = event:Find ("ids"):Child (0)
  local ids_count = ids_node.AttributesCount
  
  for i = 0, ids_count - 1 do
    local id = ids_node:Attribute (i) + 0
    
    for j = 1, #promo_actions do
      if id == promo_actions [j].unique_id then
        table.remove (promo_actions, j)
        break
      end
    end
  end
  
  if game.scripts.lobby then
    game.scripts.lobby.OnPromoActionsUpdated ()
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["select.spins"] = function (event)
  if game.scripts.level and game.scripts.level.OnSelectFreeSpins then
    game.scripts.level.OnSelectFreeSpins (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["share.post"] = function (event)
  GetSocialManager ().OnSharePostIdReceived (event:Find ("id"):Attribute (0), GetNodeChildAttribute (event, "type"))
end

HS_MESSAGE_PARSE_FUNCTIONS ["slot.fsgain"] = function (event)
  if game.scripts.level and game.scripts.level.OnFreeSpinsGain then
    game.scripts.level.OnFreeSpinsGain (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["slot.init"] = function (event)
  waiting_for_slot_init_response = false
  
  ignored_errors ["145"] = nil --disable ignoring error about slot.init after battle.interrupt
  --Do nothing
end

HS_MESSAGE_PARSE_FUNCTIONS ["slot.spin"] = function (event)
  LogFlurryEvent ("SlotSpin", {}, true)  

  if game.scripts.level and game.scripts.level.OnSlotSpinResponse then
    game.scripts.level.OnSlotSpinResponse (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["slots.update"] = function (event)
  local slots_node  = event:Find ("slots"):Child (0)
  local slots_count = slots_node.ChildrenCount

  slots = {}

  for j = 0, slots_count - 1 do
    local slot_node = slots_node:Child (j)
    
    local slot = {}
    
    slot.id        = slot_node:Find ("id"):Attribute (0) + 0
    slot.path      = slot_node:Find ("path"):Attribute (0)
    slot.game_name = slot_node:Find ("gameName"):Attribute (0)
    slot.level     = slot_node:Find ("level"):Attribute (0) + 0
    slot.lines     = slot_node:Find ("lines"):Attribute (0) + 0
    slot.voted     = GetNodeChildAttribute (slot_node, "isVote", "true") == "true"
    slot.features  = {}
    slot.data_node = slot_node
    
    local all_features = { SLOT_FEATURE_DOUBLE_GAME, SLOT_FEATURE_DOUBLE_WILD, SLOT_FEATURE_TRIPLE_WILD, SLOT_FEATURE_BONUS_ANYWHERE, SLOT_FEATURE_EXPANDED_WILD, SLOT_FEATURE_FREE_SPINS_SELECT, SLOT_FEATURE_BONUS_SPINS }
    local features     = GetNodeChildAttribute (slot_node, "feature", 0) + 0
    
    for i = 1, #all_features do
      local feature = all_features [i]
    
      if features % (2 * feature) >= feature then
        slot.features [feature] = true
      end
    end    
    
    table.insert (slots, slot)
  end

  if game.scripts.lobby and game.scripts.lobby.OnSlotsUpdated then --second check can be removed after lobby update
    game.scripts.lobby.OnSlotsUpdated ()
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["timebonus.update"] = function (event)
  local data_node = event:Find ("data"):Child (0)
  local is_new    = GetNodeChildAttribute (data_node, "isNew", "") == "true"

  if is_new and game.scripts.level and game.scripts.level.ResetAutoSpins then
    game.scripts.level.ResetAutoSpins ()
  end
  
  if game.scripts.level and game.scripts.level.OnTimeBonusUpdate then
    game.scripts.level.OnTimeBonusUpdate (event)
  elseif game.scripts.level and game.scripts.level.OnUnsupportedEventReceived then
    game.scripts.level.OnUnsupportedEventReceived (event:Find ("event"):Attribute (0), event)
  end 
end

HS_MESSAGE_PARSE_FUNCTIONS ["top.daily"] = function (event)
  top_daily_players = {}

  local players       = event:Find ("players"):Child (0)
  local players_count = players.ChildrenCount

  for i = 0, players_count - 1 do
    table.insert (top_daily_players, ParsePerson (players:Child (i)))
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["top.notification"] = function (event)
  local notification_info = {
    place     = GetNodeChildAttribute (event, "place", 100) + 0,
    top_type  = GetNodeChildAttribute (event, "topType") + 0,
    top_stage = GetNodeChildAttribute (event, "topStage", 100) + 0
  }
  
  if notification_info.top_type < 0 or notification_info.top_type > 2 then return end --ignore unknown types

  if notification_info.top_stage ~= 2 then return end --notify user only about top daily events
  
  local event_processed

  if game.scripts.level and game.scripts.level.OnTopNotification then
    event_processed = game.scripts.level.OnTopNotification (notification_info)
  end 

  if game.scripts.lobby then
    game.scripts.lobby.OnTopNotification (notification_info)
    
    event_processed = true
  end 
  
  if not event_processed then
    local pending_events = ApplicationDataGet (PENDING_TOP_NOTIFICATIONS_EVENTS) or {}
    
    table.insert (pending_events, notification_info)
    
    ApplicationDataSet (PENDING_TOP_NOTIFICATIONS_EVENTS, pending_events)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["top.users"] = function (event)
  if game.scripts.lobby then
    game.scripts.lobby.OnTopUsersReceived (event)
  end  
end

HS_MESSAGE_PARSE_FUNCTIONS ["tutorial.update"] = function (event)
  tutorials = GetNodeChildArray (event, "tutorials", {})

  if game.scripts.lobby then
    game.scripts.lobby.OnTutorialUpdate ()
  end  
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.achievements"] = function (event)
  if game.scripts.ui_windows then
    game.scripts.ui_windows.OnUserAchievementsLoaded (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.authenticate"] = function (event)
  user_id      = event:Find ("userId"):Attribute (0) + 0
  is_depositor = GetNodeChildAttribute (event, "depositor")
  
  if is_depositor then
   is_depositor = is_depositor == "true"
  end
  
  is_platform_depositor = GetNodeChildAttribute (event, "platformDep")

  if is_platform_depositor then
    is_platform_depositor = is_platform_depositor == "true"
  end
  
  authenticating = false
  
  local status = GetNodeChildAttribute (event, "status", 0) + 0
  
  if status ~= 0 and game.scripts.ui_windows then
    local texts = {}
  
    if status == 1 then
      texts = { GetLocalizedString ("SERVER_UPDATING_ERROR_LINE1"), GetLocalizedString ("SERVER_UPDATING_ERROR_LINE2"),
                GetLocalizedString ("SERVER_UPDATING_ERROR_LINE3"), GetLocalizedString ("SERVER_UPDATING_ERROR_LINE4"),
                GetLocalizedString ("SERVER_UPDATING_ERROR_LINE5") }
    else
      texts = { GetLocalizedString ("CONNECTION_ERROR"), "", "Login status " .. tostring (status) }
    end
  
    local flurry_params = { Reason = "Login status " .. tostring (status) }
  
    LogFlurryEvent ("ReconnectPopupShown", flurry_params, true)  
    
    game.scripts.ui_windows.ShowAlertWindow (nil, function ()
        LogFlurryEvent ("ReconnectPopupClick", flurry_params, true)
        HsConnectionReconnect ()
      end, nil, GetLocalizedString ("RECONNECT"), texts)

    DelayedAction ("ui", 5, function ()
      for i = 1, #stored_messages do
        local message = stored_messages [i]
      
        if message.message_id ~= DUMMY_MESSAGE then
          HsConnectionSendMessage (message.plugin, message.message)
        end
      end
    end)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.details"] = function (event)
  if game.scripts.ui_windows then
    game.scripts.ui_windows.OnUserDetailsLoaded (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.level"] = function (event)
  local levels_node    = event:Find ("levels"):Child (0)
  local previous_level = nil
  
  if user_level then
    previous_level = user_level.id
  end
  
  ParseUserLevel (levels_node:Find ("currentLevel"):Child (0), true)
  ParseUserLevel (levels_node:Find ("nextLevel"):Child (0))

  if top_daily_players then
    for i = 1, #top_daily_players do 
      local top_player = top_daily_players [i]
    
      if top_player.user_id == user_id then
        top_player.level = user_level.id + 0
      end
    end
  end

  if friends then  
    for i = 1, #friends do
      local friend = friends [i]

      if friend.user_id == user_id then
        friend.level = user_level.id + 0
      end
    end
  end

  if game.scripts.top_ui then
    game.scripts.top_ui.OnUserLevelUpdated (user_level)
  end
  
  if previous_level and user_level.id ~= previous_level and game.current_screen.name ~= "lobby" then
    local processed = false
    
    if game.scripts.level and game.scripts.level.OnUserLevelUp then
      processed = game.scripts.level.OnUserLevelUp ()
    end 
    
    if not processed then
      if game.scripts.ui_windows then
        game.scripts.ui_windows.ShowLevelUpWindow ()
      end
    end
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.ping"] = function (event)
  --Do nothing
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.profile"] = function (event)
  if game.scripts.ui_windows then
      game.scripts.ui_windows.OnUserProfileLoaded (event)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["user.update"] = function (event)
  local diamonds_node = event:Find ("diamonds")
  
  if diamonds_node then
    user_diamonds = diamonds_node:Attribute (0) + 0
  
    if game.scripts.top_ui and not ignore_next_balance_update then
      game.scripts.top_ui.OnUserDiamondsUpdated (user_diamonds)
    end
    
    if game.scripts.lobby then
      game.scripts.lobby.OnUserDiamondsUpdated (user_diamonds)
    end
  end

  local money_node = event:Find ("money")
  
  if money_node then
    user_money = money_node:Attribute (0) + 0
  
    if game.scripts.level and game.scripts.level.OnUserMoneyUpdated then
      game.scripts.level.OnUserMoneyUpdated (user_money)
    end
  
    if game.scripts.top_ui and not ignore_next_balance_update then
      game.scripts.top_ui.OnUserMoneyUpdated (user_money)
    end
  end
    
  ignore_next_balance_update = nil

  local wagering_node = event:Find ("wagering")
  
  if wagering_node then
    user_wagering = wagering_node:Attribute (0) + 0
  end

  if game.scripts.top_ui then
    game.scripts.top_ui.OnUserLevelUpdated (user_level)
  end
end

HS_MESSAGE_PARSE_FUNCTIONS ["wheel.update"] = function (event)
  wheel_data = { coins = {} }
  
  local function ReadArray (node, target_table)
    local count = node.AttributesCount

    for i = 0, count - 1 do
      table.insert (target_table, node:Attribute (i) + 0)
    end   
  end
  
  ReadArray (event:Find ("coins"):Child (0), wheel_data.coins)
end

HS_MESSAGE_PARSE_FUNCTIONS ["error"] = function (event)
  local error_id = event:Find ("errorId")
  
  if error_id then
    error_id = error_id:Attribute (0)
  end
  
  if ignored_errors [error_id] then
    ignored_errors [error_id] = nil
    print ("ERROR ignored") 
    return 
  end

  if error_id and error_id == "101" and authenticating then
    ShowUpdateVersionAlert ()
    return
  end

  if error_id and (error_id + 0 == ERROR_CODE_SLOT_LEVEL_TOO_HIGH or error_id + 0 == ERROR_CODE_SLOT_FOR_BONUS_NOT_INITIALIZED or error_id + 0 == ERROR_CODE_BONUS_NOT_ACTIVE) then
    current_slot_id = nil
    HsConnectionPopStoredMessages (AUTH_MESSAGE)
    SetCurrentScreen ("lobby")
  end 
  
  if game.scripts.ui_windows then
    DelayedAction ("ui", 1, function ()
      if #stored_messages == 1 then
        game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
          SetInputState (false)              
          HsConnectionDisconnect ()
          SetCurrentScreen ("main_menu")
        end, nil, GetLocalizedString ("CANCEL"), { GetLocalizedString ("CONNECTION_ERROR"), error_id })
      else
        local flurry_params = { Reason = "Server error " .. tostring (error_id) }
  
        LogFlurryEvent ("ReconnectPopupShown", flurry_params, true)  
    
        game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
          LogFlurryEvent ("ReconnectPopupClick", flurry_params, true)
          SetInputState (false)              
          HsConnectionPopStoredMessages (AUTH_MESSAGE)
          HsConnectionReconnect ()
        end, nil, GetLocalizedString ("RECONNECT"), { GetLocalizedString ("CONNECTION_ERROR"), error_id })
      end
    end)
  end
end

function OnHsConnectionMessage (plugin, plugin, message)
  if plugin ~= 0 and plugin ~= 1 then
    error ("Received message for unsupported plugin " .. plugin)
    return
  end

  local node         = Common.StringNode.LoadJsonFromString (message)
  local events_count = node.ChildrenCount
  
  for i = 0, events_count - 1 do
    local event = node:Child (i)
    local event_type = event:Find ("event")
    
    if event_type then
      local event_type = event_type:Attribute (0)

      local handler = HS_MESSAGE_PARSE_FUNCTIONS [event_type]
      
      if handler then
        handler (event)
      else
        print ("Received unsupported event " .. event_type)
        
        if game.scripts.level and game.scripts.level.OnUnsupportedEventReceived then
          game.scripts.level.OnUnsupportedEventReceived (event_type, event)
        end
      end
    end
  end

  for i = 0, events_count - 1 do
    local event = node:Child (i)
    local event_type = event:Find ("event")

    if event_type then
      local event_type = event_type:Attribute (0)
      
      if event_type == "slot.init" then
        HsConnectionOnStoredMessageLoaded (SLOT_INIT_MESSAGE, event)
      elseif event_type == "bonus.init" then
        HsConnectionOnStoredMessageLoaded (BONUS_INIT_MESSAGE, event)
      elseif event_type == "bonus.money" then
        HsConnectionOnStoredMessageLoaded (BONUS_MONEY_MESSAGE, event)
      elseif event_type == "user.authenticate" then
        HsConnectionOnStoredMessageLoaded (AUTH_MESSAGE, event)
      elseif event_type == "lobby.init" then
        HsConnectionOnStoredMessageLoaded ("lobby.init", event)
      end
    end
  end
end

function GetSlots ()
  return slots
end

function GetBattles ()
  return battles
end

function GetBattlesTimeout ()
  return battles_timeout
end

function GetInstallBonus ()
  return install_bonus
end

function IsDepositor ()
  return is_depositor
end

function IsPlatformDepositor ()
  return is_platform_depositor
end

function GetWelcomeBonus ()
  return welcome_bonus
end

function SetWelcomeBonus ()
  welcome_bonus = nil
end

function GetLobbyBonuses ()
  return lobby_bonuses
end

function SetInstallBonus (value)
  install_bonus = value
end

function GetUserId ()
  return user_id
end

function GetUserDiamonds ()
  return user_diamonds
end

function GetUserMoney ()
  return user_money
end

function GetUserWagering ()
  return user_wagering
end

function GetBets ()
  return bets
end

function GetUserLevel ()
  return user_level
end

function GetGrantLevel ()
  return grant_level
end

function GetGrantTime ()
  return grant_time
end

function GetGrantPeriod ()
  return grant_period
end

function GetFriends ()
  return friends
end

function GetMigrationDetails ()
  return migration_details
end

function GetSentGifts ()
  return sent_gifts
end

function GetTopDailyPlayers ()
  return top_daily_players
end

function GetTutorials ()
  return tutorials
end

function GetPayments ()
  return payments
end

function GetPaymentsDataNode ()
  return payments_data_node
end

function GetPaymentsDiamonds ()
  return payments_diamonds
end

function GetPaymentsDiamondsDataNode ()
  return payments_diamonds_data_node
end

function GetPaymentsLevelBonuses ()
  return payments_level_bonuses
end

function GetPowerupsStore ()
  return powerups_store
end

function GetPowerups ()
  return powerups
end

function GetPrizes ()
  return prizes
end

function GetPromoActions ()
  return promo_actions
end

function GetHappyTimePromoActions ()
  local actions = {}

  for i = 1, #promo_actions do
    local action = promo_actions [i]

    local additional = action.additional:Child (0)
    local type       = GetNodeChildAttribute (additional, "type", 0) + 0

    if action.id == 28 and (type > 0 and type < 6) then
      table.insert (actions, action)
    end
  end
  
  table.sort (actions, function (action_1, action_2)
    return action_1.time_left < action_2.time_left
  end)
  
  return actions
end

function GetWheelData ()
  return wheel_data
end

function GetFirstBuyBonus ()
  for i = 1, #promo_actions do
    local action = promo_actions [i]
  
    if action.type == 6 then
      if action.additional and action.additional.ChildrenCount > 0 then
        return GetNodeChildAttribute (action.additional:Child (0), "value", 100) + 0
      else
        return 100
      end
    end
  end

  return 100
end

function GetNextUserLevel ()
  if not user_level then return nil end

  return user_levels [user_level.id + 1]
end

function HsConnectionSendMessage (plugin, message)
  if message.event then
    if message.event == "slot.spin" then
      if needs_show_promo_ended_action then
        game.scripts.ui_windows.ShowHappyTimeActionEndedWindow (needs_show_promo_ended_action, function ()
          SetCurrentScreen ("lobby")
        end)
        
        return
      end
    
      if game.scripts.top_ui and game.scripts.top_ui.OnSlotSpin then
        game.scripts.top_ui.OnSlotSpin (message)
      end
    elseif message.event == "bonus.exit" then
      ApplicationDataSet (BONUS_GAME_COMPLETED, true)
    elseif message.event == "slot.init" then
      waiting_for_slot_init_response = true
  
      current_slot_id = message.slotId + 0
    
      local happy_time_actions = GetHappyTimePromoActions ()
      local promo_time_end
      local promo_duration

      for i = 1, #happy_time_actions do
        local action     = happy_time_actions [i]
        local additional = action.additional:Child (0)
        local type       = GetNodeChildAttribute (additional, "type") + 0

        if type == 5 and GetNodeChildAttribute (additional, "value1") + 0 == current_slot_id then
          promo_time_end = action.time_end
          promo_duration = GetNodeChildAttribute (additional, "duration", 1) + 0
          break
        end
      end

      local min_level

      local slots = GetSlots ()

      for i = 1, #slots do
        local slot_desc = slots [i]

        if slot_desc.id == current_slot_id then
          min_level = GetNodeChildAttribute (slot_desc.data_node, "gameLevel", 1) + 0
          break
        end
      end

      if promo_time_end then
        local time_left = math.max (0, promo_time_end - game.main_timer.GetTime ())

        promo_ended_action = DelayedAction (game.main_timer, time_left - 1, function ()
          local user_level = GetUserLevel ().id + 0

          if user_level < min_level then
            needs_show_promo_ended_action = promo_duration
          
            if game.current_screen.name ~= "loader" then
              game.scripts.ui_windows.ShowHappyTimeActionEndedWindow (promo_duration, function ()
                SetCurrentScreen ("lobby")
              end)
            end
          end
        end)
      end

      local open_doors_promo_time_end

      for i = 1, #promo_actions do
        local action = promo_actions [i]

        if action.id == 26 then
          open_doors_promo_time_end = action.time_end
          break
        end
      end

      if open_doors_promo_time_end then
        local time_left = math.max (0, open_doors_promo_time_end - game.main_timer.GetTime ())

        open_doors_promo_ended_action = DelayedAction (game.main_timer, math.max (0.01, time_left - 3), function ()
          local user_level = GetUserLevel ().id + 0

          if user_level < min_level then
            needs_show_open_doors_promo_ended_action = min_level
          
            if game.current_screen.name ~= "loader" then
              game.scripts.ui_windows.ShowOpenDoorsActionIsOver (min_level)
            end
          end
        end)
      end

      if ApplicationDataGet (IS_BATTLE_ACTIVE) then
        message.gameType = 2
      else
        message.gameType = 1
        
        local select_game_tutorial_completed = false

        for i = 1, #tutorials do
          if tutorials [i] + 0 == 1 then
            select_game_tutorial_completed = true
            break
          end
        end

        if not select_game_tutorial_completed then
          HsConnectionSendMessage (1, { event = "tutorial.complete", step = 1 })
        end
      end
    elseif message.event == "battle.search" then
      HsConnectionSendMessage (1, { event = "user.ping" }) --this message has no response, so needs to send ping with it to prevent receive timeout
    
      math.randomseed (game.main_timer.GetTime ())
    
      active_battle_search_id = math.random (10000)
      message.guid = active_battle_search_id
    elseif message.event == "battle.interrupt" then
      HsConnectionSendMessage (1, { event = "user.ping" }) --this message has no response, so needs to send ping with it to prevent receive timeout
    
      active_battle_search_id = nil
    elseif message.event == "powerup.buy" then
      local total_price = 0
    
      for i = 1, #powerups_store do
        local booster = powerups_store [i]

        if booster.id == message.id then
          total_price = booster.price * message.count
          break
        end
      end

      user_diamonds = user_diamonds - total_price
    elseif message.event == "achievement.select" then
      if top_daily_players then
        for i = 1, #top_daily_players do 
          local top_player = top_daily_players [i]
        
          if top_player.user_id == user_id then
            if message.checked then
              top_player.achiev_group_id = message.id
            else
              top_player.achiev_group_id = 0
            end
          end
        end
      end

      for i = 1, #friends do
        local friend = friends [i]
        
        if friend.user_id == user_id then
          if message.checked then
            friend.achiev_group_id = message.id
          else
            friend.achiev_group_id = 0
          end
        end
      end
      
      if game.scripts.lobby then
        game.scripts.lobby.OnPlayerSelectedAchievementUpdated ()
      end
    elseif message.event == "tutorial.complete" then
      for i = 1, #tutorials do
        if tutorials [i] + 0 == message.step then --already completed
          return
        end
      end
    
      HsConnectionSendMessage (1, { event = "user.ping" }) --this message has no response, so needs to send ping with it to prevent receive timeout
    
      table.insert (tutorials, message.step)
    elseif message.event == "double.exit" then
      if game.scripts.boosters_overlay then
        game.scripts.boosters_overlay.SetBoostersForceDisabledDuringAutospin (false)
      end
    elseif message.event == "double.init" then
      if game.scripts.boosters_overlay then
        game.scripts.boosters_overlay.SetBoostersForceDisabledDuringAutospin (true)
      end
    elseif message.event == "user.authenticate" then
      message.lang      = GetCurrentLocale ()
      message.migration = GetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE) == "true"
    elseif message.event == "payment.check" then
      is_platform_depositor = true
    end
  end
  
  if MOCKUP_ENABLED then
    print ("HsConnection SendMessage " .. plugin .. " " .. JsonCoder.encode ({ message }))
    HsMockupServerProcessMessage (plugin, message)
  else
    local message_string = JsonCoder.encode ({ message })

    message_string = string.gsub (message_string, "'", "\\u0027")

    System.Application.PostNotification ("HsConnection SendMessage " .. plugin .. " " .. plugin .. " '" .. message_string .. "'")
  end
end

function HsConnectionOnLobbyOpen ()
  if current_slot_id then
    if promo_ended_action then
      promo_ended_action.Stop ()
      promo_ended_action = nil
    end  
    
    if open_doors_promo_ended_action then
      open_doors_promo_ended_action.Stop ()
      open_doors_promo_ended_action = nil
    end  
    
    needs_show_promo_ended_action = nil
    needs_show_open_doors_promo_ended_action = nil

    HsConnectionSendMessage (1, { event = "slot.exit", slotId = current_slot_id })
    
    current_slot_id = nil
  end
end

function HsConnectionSendStoredMessage (message_id, plugin, message, reconnect_handler)
  table.insert (stored_messages, { message_id = message_id, plugin = plugin, message = message, reconnect_handler = reconnect_handler } )
  
  ClearStoredResponse (message_id)
  
  HsConnectionSendMessage (plugin, message)
end

function HsConnectionClearStoredMessages ()
  stored_responses = {}
  stored_messages  = {}
end

function HsConnectionPopStoredMessages (pop_to)
  for i = 1, #stored_messages do
    if stored_messages [i].message_id == pop_to then
      i = i + 1
    
      for j = #stored_messages, i, -1 do
        ClearStoredResponse (stored_messages [j].message_id)
        stored_messages [j] = nil
      end

      return
    end
  end
end

function HsConnectionGetStoredResponse (message_id)
  return stored_responses [message_id]
end

function HsConnectionGetLastStoredMessageId ()
  if #stored_messages == 0 then return nil end

  return stored_messages [#stored_messages].message_id
end

function HsConnectionDisconnect ()
  ClearHsConnectionData ()
  HsConnectionClearStoredMessages ()
  System.Application.PostNotification ("HsConnection Disconnect")
  connect_initiated = false
end

function HsConnectToServer ()
  local social_manager = GetSocialManager ()

  if not social_manager.IsLoggedIn () or not social_manager.IsServersSetted () then return end --connect only after logged in and servers info setted

  if connect_initiated then return end

  ClearHsConnectionData ()
    
  if MOCKUP_ENABLED then
    dofile ("data/scripts/utilities/hs_mockup_server.lua")
  else
    if DEBUG then
      print ("Connecting to " .. social_manager.GetNetworkHost () .. ":" .. social_manager.GetNetworkPort (HS_PORT))
    end
    
    connect_initiated = true
    
    System.Application.PostNotification ("HsConnection Connect " .. social_manager.GetNetworkHost () .. " " .. social_manager.GetNetworkPort (HS_PORT))
  end

  authenticating = true

  local mobile_type = 1
  local os_type     = ApplicationDataGet (OS_TYPE)
  local platform    = GetPlatform ()
    
  if os_type == "iOS" then
    if platform == "phone" then
      mobile_type = 1
    else
      mobile_type = 2
    end
  elseif os_type == "Android" then
    if GetStoreManager ().GetStoreName () == AMAZON_STORE_NAME then
      mobile_type = 7
    else
      if platform == "phone" then
        mobile_type = 3
      else
        mobile_type = 4
      end
    end
  end
  
  table.insert (stored_messages, { message_id = DUMMY_MESSAGE, plugin = 0, message = "" } )  

  social_manager.LoadUserInfo (function ()
    local system_properties = System.Application.SystemProperties
    local property_name     = "UUID"
    local device_id

    if system_properties:IsPresent (property_name) then
      device_id = system_properties:GetString (property_name)
    else
      device_id = "GuestTestId"
    end
  
    local auth_message = { event    = "user.authenticate",
                           mobile   = mobile_type,
                           lang     = GetCurrentLocale (),
                           deviceId = device_id,
                           referrer = GetProfileProperty (REFERRER_ATTRIBUTE),
                           version  = PROTOCOL_ENGINE_VERSION,
                           adid     = ApplicationDataGet (ADVERTISING_IDENTIFIER) }

    social_manager.FillAuthMessage (auth_message)

    HsConnectionSendStoredMessage (AUTH_MESSAGE, 0, auth_message, function ()
      if NeedsDownloadLobby () then
        if game.current_screen.name ~= "loader" then
          SetCurrentScreen ("loader")
        end
      else
        SetCurrentScreen ("lobby")
      end
    end)
  end)

  SetInputState (true)              
end

local function PerformReconnect ()
  SetCurrentScreen ("loader")

  ClearHsConnectionData ()

  System.Application.PostNotification ("HsConnection Reconnect")
  
  for i = 1, #stored_messages do
    local message = stored_messages [i]
      
    if message.message_id ~= DUMMY_MESSAGE then
      DelayedAction (game.main_timer, i, function ()
        HsConnectionSendMessage (message.plugin, message.message)
      end)
    end
  end
end

function HsConnectionReconnect (dont_show_alert)
  connect_initiated = true

  if not was_connected then
    if game.current_screen then
      HsConnectionDisconnect ()
      DelayedAction ("ui", 1, HsConnectToServer)
    end
     
    return 
  end

  stored_responses = {}

  reconnecting = true

  if game.scripts.ui_windows and not dont_show_alert then
    reconnect_window_shown = true
    
    game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, nil, nil, nil, { GetLocalizedString ("RECONNECTING") }, true)
      
    local reconnect_delay = 1
      
    if GetStoreManager ().GetStoreName () == AMAZON_STORE_NAME then
      local system_properties     = System.Application.SystemProperties
      local device_property_name  = "Device"

      if system_properties:IsPresent (device_property_name) and system_properties:GetString (device_property_name) == "D01E" then --First gen amazon device turns off wifi after lock
        reconnect_delay = 3
      end
    end
      
    DelayedAction ("ui", reconnect_delay, PerformReconnect)
  else
    PerformReconnect ()
  end
end

function HsConnectionOnActivate ()
  if is_active == true then return end

  is_active = true

  if disconnect_on_deactivate then
    disconnect_on_deactivate.Stop ()
  end

  if not connected and not connecting and connect_initiated then
    if game.scripts.ui_windows and game.scripts.ui_windows.IsAlertWindowShown (ALERT_TYPE_ERROR) then return end  --Do not reconnect automatically when error window shown
    
    if game.current_screen and game.current_screen.name ~= "main_menu" then
      HsConnectionReconnect ()
    end
  end
end

function HsConnectionOnDeactivate ()
  if is_active == false then return end

  is_active = false
  
  if game.scripts.ui_windows and game.scripts.ui_windows.IsAlertWindowShown (ALERT_TYPE_ERROR) then return end

  disconnect_on_deactivate = game.main_timer.DelayedAction (DISCONNECT_ON_DEACTIVATE_DELAY, function ()
    System.Application.PostNotification ("HsConnection Disconnect")
  end)
end

function OnHsConnectionError (error_code, error)
  print ("Connection error code " .. ErrorCodeName (error_code) .. " error " .. error)

  if not is_active then  --Error on android in background mode. 
    System.Application.PostNotification ("HsConnection Disconnect")
    return
  end
  
  if ignore_connection_errors then return end --ignore all errors except first one

  reconnecting = false
  reconnect_window_shown = false

  if game.current_screen and game.current_screen.name ~= "main_menu" and game.current_screen.name ~= "loader" and 
     not tried_reconnect_after_error and not ApplicationDataGet (IS_BATTLE_ACTIVE) then
    ignore_connection_errors = true
    tried_reconnect_after_error = true
    LogFlurryEvent ("AutoReconnect", {}, true)
    
    ExecuteOnNextUpdate (function ()
      ignore_connection_errors = false
      
      HsConnectionReconnect (true)
    end, true)
  else
    if game.scripts.ui_windows then
      local flurry_params = { Reason = "Connection error code " .. ErrorCodeName (error_code) .. " error " .. error }
    
      LogFlurryEvent ("ReconnectPopupShown", flurry_params, true)

      System.Application.PostNotification ("HsConnection Disconnect")

      game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
        LogFlurryEvent ("ReconnectPopupClick", flurry_params, true)
        SetInputState (false)              
        HsConnectionPopStoredMessages (AUTH_MESSAGE)
        HsConnectionReconnect ()              
      end, nil, GetLocalizedString ("RECONNECT"), { GetLocalizedString ("CONNECTION_ERROR") })
    else
      game.main_timer.DelayedAction (RECONNECT_DELAY, function ()
        HsConnectionReconnect ()
      end)
    end
  end
end

function HsReportFriendsInvited (request, friends_list)
  if #friends_list == 0 then return end

  request = "_" .. request

  local referrer_id_string = ""

  for i = 1, #friends_list do
    local prefix = ""
    
    if i > 1 then
      prefix = ","
    end
  
    referrer_id_string = referrer_id_string .. prefix .. friends_list [i] .. request  
  end

  HsConnectionSendMessage (1, { event = "user.invite", referrer_id = referrer_id_string })
end
