local AMAZON_CENTER_NAME = "AmazonCenter"
local IOS_CENTER_NAME    = "IOSCenter"

local function OnRegisterFinished (success, status, token)
  if not success then
    print ("Push registration failed, status " .. tostring (status))
    ApplicationDataSet (REGISTERED_FOR_PUSH_NOTIFICATIONS, nil)
    return
  end
    
  if DEBUG then
    print ("LUA received push token '" .. token .. "'")
  end

  if not ApplicationDataGet (REGISTERED_FOR_PUSH_NOTIFICATIONS) then return end

  local hs_port = GetHsPort ()
  local url     = ApplicationDataGet (PUSH_NOTIFICATIONS_SERVER_URL) .. "online?app=slots&token=" .. token .. "&locale=" .. string.sub (GetCurrentLocale (), 1, 2) .."&userid=" .. GetUserId () .. "&network=" .. GetSocialManager ().GetNetworkName () .. "&env="

  if hs_port == HS_LIVE_PORT then
    url = url .. "live"
  elseif hs_port == HS_QA_PORT then
    url = url .. "qa"
  else
    url = url .. "dev"
  end

  local system_properties = System.Application.SystemProperties
  local uid_property_name = "UUID"

  if system_properties:IsPresent (uid_property_name) then
    url = url .. "&uid=" .. system_properties:GetString (uid_property_name)
  end

  local platform_code = "iphone"

  if ApplicationDataGet (OS_TYPE) == "Android" then
    if GetPushNotificationsManager ().GetCenterName () == AMAZON_CENTER_NAME then
      platform_code = "amazon"
    else
      platform_code = "android"
    end
  elseif GetPlatform () == "pad" then
    platform_code = "ipad"
  end

  url = url .. "&platform=" .. platform_code

  if DEBUG then
    print ("Registering for push notification with url '" .. url .. "'")
  end

  Common.File.AsyncLoadString (url, Common.File.CreateAsyncLoadHandler (function (result)
    if DEBUG then
      print ("Push registration result '" .. result .. "'")
    end
  end))
end

local function RegisterForPushNotifications (impl)
  if not impl.center then return end
  
  if ApplicationDataGet (REGISTERED_FOR_PUSH_NOTIFICATIONS) then return end

  ApplicationDataSet (REGISTERED_FOR_PUSH_NOTIFICATIONS, true)

  local register_properties = Common.PropertyMap.Create ()

  if impl.center_name == IOS_CENTER_NAME then
    register_properties:SetString ("Types", "Badge Sound Alert")
  end
  
  impl.center:RegisterForNotifications (PushNotifications.Center.CreateRegisterCallback (OnRegisterFinished), register_properties)
end

local function OnPushNotificationReceived (impl, notification)
  if DEBUG then
    print ("push notification received '" .. notification .. "'")
  end

  if impl.center_name == IOS_CENTER_NAME then
    local node                  = Common.StringNode.LoadJsonFromString (notification)
    local received_while_active = GetNodeChildAttribute (node, "received_while_active", 0) + 0

    if received_while_active ~= 0 then
      local aps_node   = node:Find ("aps")
      local alert_node = aps_node:Child (0):Find ("alert")
    
      if alert_node and alert_node.AttributesCount > 0 then
        local alert = alert_node:Attribute (0)
      
        System.Application.PostNotification ("ShowSystemAlert push_alert|" .. alert .. "| |" .. GetLocalizedString ("CLOSE"))
      end
    end

    System.Application.SystemProperties:SetInteger ("ApplicationIconBadgeNumber", 0)
  end
end

local function CreatePushNotificationsManager ()
  local impl = {}
  
  local centers_names = {}

  local os = ApplicationDataGet (OS_TYPE)

  if string.find (os, "iOS") ~= nil then
    table.insert (centers_names, IOS_CENTER_NAME)
  elseif os == "Android" then
    local system_properties = System.Application.SystemProperties

    table.insert (centers_names, "GoogleCenter")

    if system_properties:GetString ("Manufacturer") == "Amazon" then --amazon center initialized after google center in case of google play devices with 'Amazon' manufacturer
      table.insert (centers_names, AMAZON_CENTER_NAME)
    end
  end

  for i = 1, #centers_names do
    local center_name = centers_names [i]
      
    print ("Creating push notifications center " .. center_name)
      
    local status, err = pcall (function ()
      impl.center      = PushNotifications.Center.Create (center_name)
      impl.center_name = center_name 
      
      impl.center:RegisterNotificationsHandler (PushNotifications.Center.CreateNotificationsHandler (function (notification)
        OnPushNotificationReceived (impl, notification)
      end))
    end)
    
    if status then break end --success
    
    print ("Can't create push notifications center " .. center_name .. ", error " .. err)
  end

  local public_methods = {}

  public_methods.GetCenterName                = function () return impl.center_name end
  public_methods.RegisterForPushNotifications = function () RegisterForPushNotifications (impl) end
    
  return public_methods
end

local PushNotificationsManagerInstance

function GetPushNotificationsManager ()
  if not PushNotificationsManagerInstance then
    PushNotificationsManagerInstance = CreatePushNotificationsManager ()
  end
  
  return PushNotificationsManagerInstance
end
