local ACCELERATION_PER_PIXEL   = 5    -- ускорение при прокрутке на 1 пиксел
local ACCELERATION_STOP_TIME   = 0.1  -- время, по истечении которого ускорение обнуляется, если палец не был отпущен 
local ANIMATED_SCROLL_DURATION = 0.3
local CELL_RELEASE_DRAG        = 4
local DECELERATION_TIME        = 2    -- время замедления
local EPS                      = 0.0001
local RETURN_TIME              = 0.5  -- время возврата при перескроллировании
local SPEED_CALC_BUFFER        = 0.2  -- we will keep move events for last SPEED_CALC_BUFFER seconds
local SPEED_STOP_THRESHOLD     = 300  -- disable decelerated scrolling on touch end if scrolling speed is less then threshold
local TABLE_SCROLL_EPS         = 0.04

local function CancelAnimation (impl)
  if impl.animated_scroll_action then
    impl.animated_scroll_action.Stop ()
    impl.animated_scroll_action = nil
  end
  
  impl.deceleration_time = 0
  impl.return_time.x = 0
  impl.return_time.y = 0
end

local function OnBeginTouch (impl, x, y)
  CancelAnimation (impl)

  impl.touched = true
  
  impl.last_x = x
  impl.last_y = y
  impl.new_x  = x
  impl.new_y  = y
  
  impl.last_input_events = {}

  if impl.table_delegate and impl.cell_size then
    local position   = impl.node.Position
    local cell_index = nil
  
    local x_click_offset = x - position.x
    local y_click_offset = y - position.y    
  
    if impl.horizontal_scroll_enabled then
      cell_index = math.ceil (x_click_offset / impl.cell_size)
    elseif impl.vertical_scroll_enabled then
      cell_index = math.ceil (y_click_offset / impl.cell_size + 0.5)
    end

    if cell_index > 0 then
      local cell_x_offset = x_click_offset
      local cell_y_offset = y_click_offset
      
      if impl.horizontal_scroll_enabled then
        cell_x_offset = cell_x_offset % math.abs (impl.cell_size)
      elseif impl.vertical_scroll_enabled then
        cell_y_offset = (cell_y_offset + impl.cell_size * 0.5) % math.abs (impl.cell_size)
      end      
    
      impl.pressed_cell = { index = cell_index, x = cell_x_offset, y = cell_y_offset }

      if impl.table_delegate.OnCellPressed then
        local status, err = pcall (function ()
          impl.table_delegate.OnCellPressed (impl.pressed_cell.index, impl.pressed_cell.x, impl.pressed_cell.y)
        end)

        if not status then
          print (err)
        end
      end
    end
  end
end

local function OnEndTouch (impl)
  if not impl.touched then return end

  impl.needs_process_touch_end = true
  impl.touched = false
end

local function OnDrag (impl, x, y)
  if not impl.touched then return end

  impl.new_x = x
  impl.new_y = y
end

local function Update (impl, dt)
  local dx = impl.new_x - impl.last_x
  local dy = impl.new_y - impl.last_y

  impl.last_x = impl.new_x
  impl.last_y = impl.new_y

  if not impl.horizontal_scroll_enabled then
    dx = 0 
  end

  if not impl.vertical_scroll_enabled then
    dy = 0 
  end

  for i = #impl.last_input_events, 1, -1 do
    impl.last_input_events [i].time_to_live = impl.last_input_events [i].time_to_live - dt
    
    if i ~= 1 and impl.last_input_events [i].time_to_live < 0 then --dont erase last event for low fps situations
      table.remove (impl.last_input_events, i)
    end
  end

  table.insert (impl.last_input_events, 1, { dx = dx, dy = dy, time_to_live = SPEED_CALC_BUFFER, dt = dt })

  if dx * impl.acceleration.x < 0 then
    impl.acceleration.x = 0
  end

  if dy * impl.acceleration.y < 0 then
    impl.acceleration.y = 0
  end

  if (dx ~= 0 or dy ~= 0) and (impl.touched or impl.needs_process_touch_end) and impl.pressed_cell then
    if math.abs (dx) > CELL_RELEASE_DRAG or math.abs (dy) > CELL_RELEASE_DRAG then
      if impl.table_delegate.OnCellTouchCancelled then
        local status, err = pcall (function ()
          impl.table_delegate.OnCellTouchCancelled (impl.pressed_cell.index, impl.pressed_cell.x, impl.pressed_cell.y)
        end)

        if not status then
          print (err)
        end
      end

      impl.pressed_cell = nil
    end
  end 

  impl.acceleration_stop_time = math.max (0, impl.acceleration_stop_time - dt)

  if dx ~= 0 or dy ~= 0 then 
    impl.node:Translate (dx, dy, 0)

    impl.acceleration.x         = impl.acceleration.x + dx * ACCELERATION_PER_PIXEL
    impl.acceleration.y         = impl.acceleration.y + dy * ACCELERATION_PER_PIXEL
    impl.acceleration_stop_time = ACCELERATION_STOP_TIME

    if impl.scroll_callback then
      impl.scroll_callback ()
    end
  end
  
  if impl.acceleration_stop_time < 0 then
    impl.acceleration_stop_time = 0
    impl.acceleration.x         = 0
    impl.acceleration.y         = 0
  end

  if not impl.touched then
    if impl.needs_process_touch_end then
      impl.needs_process_touch_end = false
      
      local total_dx = 0 
      local total_dy = 0 
      local total_dt = 0

      for i = 1, #impl.last_input_events do
        local entry = impl.last_input_events [i]

        total_dx = total_dx + entry.dx
        total_dy = total_dy + entry.dy
        total_dt = total_dt + entry.dt
      end
      
      impl.last_input_events = {}

      local speed_x = math.abs (total_dx / total_dt)
      local speed_y = math.abs (total_dy / total_dt)
     
      impl.back_start_position = { x = nil, y = nil }

      impl.deceleration_time = DECELERATION_TIME

      if speed_x < SPEED_STOP_THRESHOLD and speed_y < SPEED_STOP_THRESHOLD then
        impl.deceleration_time = 0
      
        if impl.pressed_cell then
          if impl.table_delegate.OnCellClick then
            local status, err = pcall (function ()
              impl.table_delegate.OnCellClick (impl.pressed_cell.index, impl.pressed_cell.x, impl.pressed_cell.y)
            end)

            if not status then
              print (err)
            end
          end
        end
      end

      if impl.pressed_cell then
        impl.pressed_cell = nil
      end
    end

    impl.deceleration_time = impl.deceleration_time - dt
    
    if impl.deceleration_time < 0 then
      impl.deceleration_time = 0
    end
    
    local acceleration_value = impl.deceleration_time / DECELERATION_TIME
    local acceleration_x     = impl.acceleration.x * acceleration_value
    local acceleration_y     = impl.acceleration.y * acceleration_value
    
    if not impl.horizontal_scroll_enabled then
      acceleration_x = 0
    end

    if not impl.vertical_scroll_enabled then
      acceleration_y = 0
    end

    local scrolled = false

    if acceleration_x ~= 0 or acceleration_y ~= 0 then
      impl.node:Translate (acceleration_x * dt, acceleration_y * dt, 0)
      scrolled = true
    end
    
    impl.return_time.x = impl.return_time.x - dt
    
    if impl.return_time.x < 0 then
      impl.return_time.x = 0
    end
    
    impl.return_time.y = impl.return_time.y - dt
    
    if impl.return_time.y < 0 then
      impl.return_time.y = 0
    end
    
    local position = impl.node.Position

    if impl.back_start_position.x then
      local return_value = impl.return_time.x / RETURN_TIME
      
      if impl.back_start_position.x < impl.minimum_position.x then
        position.x = impl.minimum_position.x - (impl.minimum_position.x - impl.back_start_position.x) * return_value
      end 
      if impl.back_start_position.x > impl.maximum_position.x then
        position.x = impl.maximum_position.x + (impl.back_start_position.x - impl.maximum_position.x) * return_value 
      end 

      scrolled = true
      
      if impl.return_time.x == 0 then
        impl.back_start_position.x = nil
      end
    else
      if position.x < impl.minimum_position.x or position.x > impl.maximum_position.x then
        impl.back_start_position.x = position.x
        impl.acceleration.x        = 0
      end

      impl.return_time.x = RETURN_TIME
    end 

    if impl.back_start_position.y then
      local return_value = impl.return_time.y / RETURN_TIME
      
      if impl.back_start_position.y < impl.minimum_position.y then
        position.y = impl.minimum_position.y - (impl.minimum_position.y - impl.back_start_position.y) * return_value 
      end 
      if impl.back_start_position.y > impl.maximum_position.y then
        position.y = impl.maximum_position.y + (impl.back_start_position.y - impl.maximum_position.y) * return_value 
      end 

      scrolled = true
      
      if impl.return_time.y == 0 then
        impl.back_start_position.y = nil
      end
    else
      if position.y < impl.minimum_position.y or position.y > impl.maximum_position.y then
        impl.back_start_position.y = position.y
        impl.acceleration.y        = 0
      end

      impl.return_time.y = RETURN_TIME
    end 
    
    if scrolled then
      impl.node.Position = position
  
      if impl.scroll_callback then
        impl.scroll_callback ()
      end
    end
  end
end

local function ScrollToRelative (impl, x, y, animated)
  CancelAnimation (impl)  --cancel current scrolling

  impl.last_x = impl.new_x   
  impl.last_y = impl.new_y
  impl.last_input_events = {}
  impl.back_start_position = { x = nil, y = nil }

  local start_position  = impl.node.Position
  local target_position = vec3 (impl.minimum_position.x + (impl.maximum_position.x - impl.minimum_position.x) * x, impl.minimum_position.y + (impl.maximum_position.y - impl.minimum_position.y) * y, start_position.z)
  
  if animated then
    local position_diff = target_position - start_position
  
    impl.animated_scroll_action = SlowAction ("ui", ANIMATED_SCROLL_DURATION, function (t)
      impl.node.Position = start_position + position_diff * t
      
      if impl.scroll_callback then
        impl.scroll_callback ()
      end
    end)
  else
    impl.node.Position = target_position
  
    if impl.scroll_callback then
      impl.scroll_callback ()
    end
  end 
end

local function SetScrollEnabled (impl, horizontal, vertical)
  impl.horizontal_scroll_enabled = horizontal
  impl.vertical_scroll_enabled   = vertical
end

local function ScrollPosition (impl)
  local position     = impl.node.Position
 
  return { x = position.x - impl.minimum_position.x, y = position.y - impl.minimum_position.y }
end

local function RelativeScrollPosition (impl, clamped)
  local position     = impl.node.Position
  local return_value = { x = 0, y = 0 }
  
  if math.abs (impl.maximum_position.x - impl.minimum_position.x) > 0 then
    return_value.x = (position.x - impl.minimum_position.x) / (impl.maximum_position.x - impl.minimum_position.x)
  end
  
  if math.abs (impl.maximum_position.y - impl.minimum_position.y) > 0 then
    return_value.y = (position.y - impl.minimum_position.y) / (impl.maximum_position.y - impl.minimum_position.y)
  end
    
  if clamped then
    return_value.x = math.max (math.min (return_value.x, 1), 0)
    return_value.y = math.max (math.min (return_value.y, 1), 0)
  end
  
  return return_value
end

local function GetFirstVisibleCell (impl)
  if not impl.cell_size then return 0 end

  local position = impl.node.Position
  local return_value = 0
  
  if impl.horizontal_scroll_enabled then
    return_value = (impl.maximum_position.x - position.x) / impl.cell_size
  elseif impl.vertical_scroll_enabled then
    return_value = (impl.minimum_position.y - position.y) / impl.cell_size
  end
  
  if math.abs (return_value - math.ceil (return_value)) < TABLE_SCROLL_EPS then
    return_value = math.ceil (return_value)
  end
  
  if math.abs (return_value - math.floor (return_value)) < TABLE_SCROLL_EPS then
    return_value = math.floor (return_value)
  end
  
  return return_value
end

local function GetLastVisibleCell (impl)
  if not impl.cell_size or not impl.table_size then return 0 end

  local position = impl.node.Position
  local return_value = 0
  
  if impl.horizontal_scroll_enabled then
    return_value = (impl.maximum_position.x - position.x + impl.table_size) / impl.cell_size
  elseif impl.vertical_scroll_enabled then
    return_value = (impl.minimum_position.y - position.y - impl.table_size) / impl.cell_size
  end
  
  if math.abs (return_value - math.ceil (return_value)) < TABLE_SCROLL_EPS then
    return_value = math.ceil (return_value)
  end
  
  if math.abs (return_value - math.floor (return_value)) < TABLE_SCROLL_EPS then
    return_value = math.floor (return_value)
  end
  
  return return_value
end

local function GetCellsCount (impl)
  if not impl.cell_size or not impl.table_size then return 0 end

  local return_value = 0
  
  if impl.horizontal_scroll_enabled then
    return_value = (math.abs (impl.maximum_position.x - impl.minimum_position.x) + impl.table_size) / impl.cell_size
  elseif impl.vertical_scroll_enabled then
    return_value = (math.abs (impl.maximum_position.y - impl.minimum_position.y) + impl.table_size) / impl.cell_size
  end
  
  if math.abs (return_value - math.ceil (return_value)) < TABLE_SCROLL_EPS then
    return_value = math.ceil (return_value)
  end
  
  if math.abs (return_value - math.floor (return_value)) < TABLE_SCROLL_EPS then
    return_value = math.floor (return_value)
  end
  
  return return_value
end

local function ScrollToCell (impl, cell, animated)
  if not impl.cell_size then return 0 end

  local scroll_offset = impl.cell_size * cell

  if impl.horizontal_scroll_enabled then
    scroll_offset = 1 - math.max (0, math.min (1, scroll_offset / (impl.maximum_position.x - impl.minimum_position.x)))
  
    ScrollToRelative (impl, scroll_offset, 0, animated)
  elseif impl.vertical_scroll_enabled then
    scroll_offset = math.max (0, math.min (1, scroll_offset / (impl.maximum_position.y - impl.minimum_position.y)))
  
    ScrollToRelative (impl, 0, scroll_offset, animated)
  end
end

local function SetScrollRange (impl, x, y) 
  impl.minimum_position = impl.initial_position - vec3 (x or 0, 0, 0)
  impl.maximum_position = impl.minimum_position + vec3 (x or 0, y or 0, 0)
end

-- Создание Scroll View
function CreateScrollView (in_node, in_horizontal_scroll_enabled, in_vertical_scroll_enabled, scroll_range_x, scroll_range_y, zone_sprite, scrollers_config)
  local impl = {}

  impl.node                      = in_node
  impl.horizontal_scroll_enabled = in_horizontal_scroll_enabled
  impl.vertical_scroll_enabled   = in_vertical_scroll_enabled
  impl.initial_position          = in_node.Position
  impl.minimum_position          = impl.initial_position - vec3 (scroll_range_x or 0, 0, 0)
  impl.maximum_position          = impl.minimum_position + vec3 (scroll_range_x or 0, scroll_range_y or 0, 0)
  impl.acceleration              = { x = 0, y = 0 }
  impl.acceleration_stop_time    = 0
  impl.deceleration_time         = 0
  impl.return_time               = { x = 0, y = 0 }
  impl.back_start_position       = { x = nil, y = nil }
  
  impl.new_x  = 0
  impl.last_x = 0
  impl.new_y  = 0
  impl.last_y = 0

  impl.last_input_events = {}

  if zone_sprite then
    impl.zone_sprite   = GetSprite (zone_sprite)
    impl.zone_handlers = {}
    
    table.insert (impl.zone_handlers, AttachInputZoneToNode (impl.zone_sprite.node, Scene.InputZoneNotification.OnPress, function (viewport, notification, context)
      OnBeginTouch (impl, context.TouchWorldPosition.x, context.TouchWorldPosition.y)
    end))
    
    table.insert (impl.zone_handlers, AttachInputZoneToNode (impl.zone_sprite.node, Scene.InputZoneNotification.OnScreenTouchUp, function (viewport, notification, context)
      OnEndTouch (impl)
    end))
    
    table.insert (impl.zone_handlers, AttachInputZoneToNode (impl.zone_sprite.node, Scene.InputZoneNotification.OnPressMove, function (viewport, notification, context)
      OnDrag (impl, context.TouchWorldPosition.x, context.TouchWorldPosition.y)
    end))
  end
  
  if scrollers_config then
    impl.scrollers = {}
    
    local function CreateScroller (desc, x_factor, y_factor)
      local zone_layout = GetSprite (desc.zone).layout
      local zone_left   = zone_layout:Get ("Left") + 0     
      local zone_width  = zone_layout:Get ("Width") + 0     
      local zone_top    = zone_layout:Get ("Top") + 0     
      local zone_height = zone_layout:Get ("Height") + 0     
    
      local scroller = {}
      
      scroller.sprite         = GetSprite (desc.sprite)
      scroller.pressed_sprite = GetSprite (desc.pressed_sprite)
      
      scroller.handlers = {}
      
      table.insert (scroller.handlers, AttachInputZoneToNode (scroller.sprite.node, Scene.InputZoneNotification.OnPress, function (viewport, notification, context)
        scroller.scroll_started = true
        
        ActivateSprite (scroller.pressed_sprite)
      end))

      table.insert (scroller.handlers, AttachInputZoneToNode (scroller.sprite.node, Scene.InputZoneNotification.OnScreenTouchUp, function (viewport, notification, context)
        scroller.scroll_started = false
        
        DeactivateSprite (scroller.pressed_sprite)
      end))

      table.insert (scroller.handlers, AttachInputZoneToNode (scroller.sprite.node, Scene.InputZoneNotification.OnScreenTouchMove, function (viewport, notification, context)
        if not scroller.scroll_started then return end
        
        local x_offset = math.min (1, math.max (0, (context.TouchWorldPosition.x - zone_left) / zone_width)) * x_factor
        local y_offset = math.min (1, math.max (0, (-zone_top - context.TouchWorldPosition.y) / zone_height)) * y_factor

        ScrollToRelative (impl, x_offset, y_offset)
      end))
      
      return scroller
    end
  
    if scrollers_config.horizontal then
      impl.scrollers.horizontal = CreateScroller (scrollers_config.horizontal, 1, 0)
    end

    if scrollers_config.vertical then
      impl.scrollers.vertical = CreateScroller (scrollers_config.vertical, 0, 1)
    end
  end

  impl.node:RegisterEventHandler (Scene.NodeEvent.BeforeSceneDetach, Scene.Node.CreateEventHandler (function ()
    impl.scroll_callback = nil
    impl.table_delegate  = nil
  
    if impl.zone_handlers then
      for i = 1, #impl.zone_handlers do
        impl.zone_handlers [i].connection:Disconnect ()
        impl.zone_handlers [i].callback = nil
      end      
    end
    
    if impl.scrollers then
      local function CleanupScroller (scroller)
        for i = 1, #scroller.handlers do
          scroller.handlers [i].connection:Disconnect ()
          scroller.handlers [i].callback = nil
        end
      end
    
      if impl.scrollers.horizontal then
        CleanupScroller (impl.scrollers.horizontal)
      end
    
      if impl.scrollers.vertical then
        CleanupScroller (impl.scrollers.vertical)
      end
    end
  end))

  local methods = {}

  methods.GetCellsCount             = function ()         return GetCellsCount (impl) end
  methods.GetFirstVisibleCell       = function ()         return GetFirstVisibleCell (impl) end
  methods.GetLastVisibleCell        = function ()         return GetLastVisibleCell (impl) end
  methods.IsHorizontalScrollEnabled = function ()         return impl.horizontal_scroll_enabled end  
  methods.IsVerticalScrollEnabled   = function ()         return impl.vertical_scroll_enabled end
  methods.SetMinimumX               = function (x)        impl.minimum_position.x = x end
  methods.SetMaximumX               = function (x)        impl.maximum_position.x = x end
  methods.SetMinimumY               = function (y)        impl.minimum_position.y = y end
  methods.SetMaximumY               = function (y)        impl.maximum_position.y = y end
  methods.GetMinimumX               = function ()         return impl.minimum_position.x end
  methods.GetMaximumX               = function ()         return impl.maximum_position.x end
  methods.GetMinimumY               = function ()         return impl.minimum_position.y end
  methods.GetMaximumY               = function ()         return impl.maximum_position.y end
  methods.OnBeginTouch              = function (x, y)     OnBeginTouch     (impl, x, y) end
  methods.OnDrag                    = function (x, y)     OnDrag           (impl, x, y) end
  methods.OnEndTouch                = function ()         OnEndTouch       (impl) end
  methods.Update                    = function (dt)       Update           (impl, dt) end
  methods.ScrollPosition            = function ()         return ScrollPosition (impl) end
  methods.ScrollToRelative          = function (x, y)     ScrollToRelative (impl, x, y) end
  methods.ScrollToCell              = function (cell, animated) ScrollToCell (impl, cell, animated) end
  methods.SetTableParams            = function (cell_size, table_size) impl.cell_size = cell_size; impl.table_size = table_size end
  methods.SetScrollCallback         = function (callback) impl.scroll_callback = callback end
  methods.SetScrollEnabled          = function (h, v)     SetScrollEnabled (impl, h, v) end
  methods.SetScrollRange            = function (x, y)     SetScrollRange (impl, x, y) end
  methods.SetTableDelegate          = function (delegate) impl.table_delegate = delegate end
  methods.RelativeScrollPosition    = function (clamped)  return RelativeScrollPosition (impl, clamped) end

  return methods
end
