local DEFAULT_SPINUP_TIME         = 0.5
local EPS                         = 0.0001
local BONUS_TURBO_SPIN_STOP_DELAY = 2

local function SetReelsScissor (impl, scissor) 
  impl.reels_scissor = scissor

  for i = 1, #impl.reels do
    local reel = impl.reels [i]
    
    if reel.items then
      for j = 1, #reel.items do
        local item = reel.items [j]

        item.sprite.Scissor        = scissor
        item.blured_sprite.Scissor = scissor
      end
    end
  end
end

local function StopSpin (impl, stop_positions, stop_sounds)
  impl.needs_stop_spinning = true
  
  impl.stop_sounds = stop_sounds

  for i = 1, math.min (#stop_positions, #impl.reels) do
    impl.reels [i].stop_position = stop_positions [i]
  end
end

local function StartSpin (impl, spin_sound_name, fast_mode)
  if impl.spin_started then return end

  impl.turbo_mode_activated = false

  if fast_mode then
    impl.spin_start_delay         = impl.fast_spin_start_delay
    impl.spin_speed               = impl.fast_spin_speed
    impl.spin_stop_delay          = impl.fast_spin_stop_delay
    impl.spin_delay               = impl.fast_spin_delay
    impl.reel_spinup_acceleration = impl.fast_reel_spinup_acceleration
    impl.spinup_time              = impl.fast_spinup_time
    impl.spin_min_duration        = impl.fast_spin_min_duration
    impl.reel_spinup_distance     = impl.fast_reel_spinup_distance
  else
    impl.spin_start_delay         = impl.normal_spin_start_delay
    impl.spin_speed               = impl.normal_spin_speed
    impl.spin_stop_delay          = impl.normal_spin_stop_delay
    impl.spin_delay               = impl.normal_spin_delay
    impl.reel_spinup_acceleration = impl.normal_reel_spinup_acceleration
    impl.spinup_time              = impl.normal_spinup_time
    impl.spin_min_duration        = impl.normal_spin_min_duration
    impl.reel_spinup_distance     = impl.normal_reel_spinup_distance
  end

  impl.spin_started        = true
  impl.spin_start_time     = game.current_screen.timers.default:GetTime ()
  impl.spin_stopping       = false
  impl.needs_stop_spinning = false
  
  for i = 1, #impl.reels do
    if impl.spinning_directions [i] ~= nil then
      local reel = impl.reels [i]

      reel.spinning             = true  
      reel.items_alpha          = -1
      reel.spin_start_position  = reel.node.Position.y - reel.start_position.y
      reel.spin_stop_start_time = nil
    end
  end

  if spin_sound_name then
    impl.spin_sound_name = spin_sound_name
    PlayStoredSound (impl.spin_sound_name)
  end
end

local function ClampReelPosition (reel, position)
  return position % reel.length
end

local function ClampReelIndex (reel, index)
  local reel_size = #reel.items

  return (index - 1) % reel_size + 1
end

local function HideInvisibleItems (impl, reel, offset, visible_alpha)
  if not impl.visible_lines_count then return end

  offset = offset or 0

  local reel_position    = reel.node.Position.y - reel.start_position.y
  local top_visible_item = math.floor (reel_position / impl.line_interval) + 1
  local hide_item        = top_visible_item + impl.visible_lines_count + offset
    
  for i = top_visible_item, hide_item - 1 do
    if i <= #reel.items then
      reel.items [i].sprite.Alpha        = visible_alpha
      reel.items [i].blured_sprite.Alpha = 1 - visible_alpha
    end
  end
    
  if hide_item <= #reel.items then
    reel.items [hide_item].sprite.Alpha        = 0
    reel.items [hide_item].blured_sprite.Alpha = 0
  end
end

local function GetLastUnlockedReel (impl)
  for i = #impl.reels, 1, -1 do
    if impl.spinning_directions [i] ~= nil then
      return i
    end
  end
end

local function PrepareReelForStopping (impl, reel_index)
  local reel                 = impl.reels [reel_index]
  local reel_position        = reel.node.Position.y - reel.start_position.y
  local stop_node            = reel.stop_node
  local top_visible_item     = math.floor (reel_position / impl.line_interval) + 1
  local first_fake_reel_item = reel.stop_position - 1  
  local last_fake_reel_item  = reel.stop_position + 2  

  local stop_items_indices = {}
  local first_item_offset  = -2

  if impl.spinning_directions [reel_index] then
    for i = first_fake_reel_item, last_fake_reel_item do
      table.insert (stop_items_indices, i)
    end  

    for i = top_visible_item, top_visible_item + 3 do
      table.insert (stop_items_indices, i)
    end
  else  
    first_item_offset = 2
  
    for i = top_visible_item, top_visible_item + 3 do
      table.insert (stop_items_indices, i)
    end

    first_fake_reel_item = first_fake_reel_item + 1 
    last_fake_reel_item  = last_fake_reel_item + 1

    for i = first_fake_reel_item, last_fake_reel_item do
      table.insert (stop_items_indices, i)
    end  
  end

  reel.stop_items = {}

  for i = 1, #stop_items_indices do
    local item = {}

    item.desc = reel.items [ClampReelIndex (reel, stop_items_indices [i])].desc

    item.sprite = Scene.Sprite.Create ()

    item.sprite.Material   = item.desc.material
    item.sprite.TexSize    = item.desc.tex_size
    item.sprite.TexOffset  = item.desc.tex_offset
    item.sprite.Position   = vec3 (0, -(i + first_item_offset) * impl.line_interval, 0)
    item.sprite.WorldScale = item.desc.scale
    item.sprite.Alpha      = 0
    item.sprite.Scissor    = impl.reels_scissor

    item.sprite:BindToParent (stop_node)

    item.blured_sprite = Scene.Sprite.Create ()

    item.blured_sprite.Material   = item.desc.blured_material
    item.blured_sprite.TexSize    = item.desc.blured_tex_size
    item.blured_sprite.TexOffset  = item.desc.blured_tex_offset
    item.blured_sprite.Position   = vec3 (0, -(i + first_item_offset) * impl.line_interval, -EPS)
    item.blured_sprite.WorldScale = item.desc.blured_scale
    item.blured_sprite.Scissor    = impl.reels_scissor

    item.blured_sprite:BindToParent (stop_node)
    
    table.insert (reel.stop_items, item)
  end

  reel.node:Unbind ()
  stop_node:BindToParent (game.current_screen.node)

  reel.stop_start_offset = reel_position % impl.line_interval

  stop_node.Position = reel.start_position + vec3 (0, reel.stop_start_offset + (#reel.stop_items - 5) * impl.line_interval, 0)
  
  if impl.spinning_directions [reel_index] then
    reel.stop_length = reel.stop_start_offset + (#reel.stop_items - 5) * impl.line_interval
  else
    reel.stop_length = impl.line_interval - reel_position % impl.line_interval + (#reel.stop_items - 5) * impl.line_interval
  end
  
  reel.stop_start_position   = stop_node.Position
  reel.stop_items_alpha      = 0
  reel.stopping_started      = false

  if reel_index < GetLastUnlockedReel (impl) then
    local next_unlocked_reel = reel_index + 1
    
    for i = next_unlocked_reel, #impl.reels do
      if impl.spinning_directions [i] ~= nil then
        next_unlocked_reel = i
        break
      end
    end
  
    DelayedAction ("default", impl.spin_stop_delay, function ()
      impl.reels [next_unlocked_reel].spin_stop_start_time = game.current_screen.timers.default:GetTime ()
      PrepareReelForStopping (impl, next_unlocked_reel)            
    end)

    if impl.event_handler and impl.event_handler.OnReelStopInitiated then
      impl.event_handler.OnReelStopInitiated (next_unlocked_reel)
    end
  end
end

local function UpdateReelGroups (impl, reel)
  local reel_position = reel.node.Position.y

  for i = 1, #reel.items_groups do
    local group = reel.items_groups [i]
    
    if group.min_y <= reel_position and group.max_y >= reel_position then
      if not group.binded then
        group.node:BindToParent (reel.node)
        group.binded = true
      end
    else
      if group.binded then
        group.node:Unbind ()
        group.binded = false
      end
    end
  end
end

local function UpdateReelSpin (impl, reel_index, spin_time)
  local reel = impl.reels [reel_index]
  
  if not reel.spinning then return end
  
  local node                 = reel.node
  local stop_node            = reel.stop_node
  local reel_time            = spin_time - (reel_index - 1) * impl.spin_start_delay  
  local full_speed_reel_time = reel_time - impl.reel_spinup_time
  local blured_alpha         = math.min (math.max (0, reel_time / impl.blur_show_time), 1)
  local alpha                = 1 - blured_alpha
  local reel_position        = 0

  local acceleration    = impl.reel_spinup_acceleration
  local speed           = impl.spin_speed
  local spinup_distance = impl.reel_spinup_distance
  local spline_factor   = 1

  if not impl.spinning_directions [reel_index] then
    acceleration    = -acceleration
    speed           = -speed
    spinup_distance = -spinup_distance
    spline_factor   = -1
  end

  if spin_time < impl.spinup_time then  --spinning up
    if reel_time > 0 then
      if reel_time < impl.reel_spinup_time then
        reel_position = reel_time * reel_time * acceleration / 2
      else
        reel_position = spinup_distance + full_speed_reel_time * speed 
      end
    end
  elseif spin_time < impl.spin_min_duration then  --spinning
    reel_position = spinup_distance + full_speed_reel_time * speed 
  else
    if reel.spin_stop_start_time then
      local reel_stop_time  = game.current_screen.timers.default:GetTime () - reel.spin_stop_start_time
      local total_stop_time = math.abs (reel.stop_length / speed)

      blured_alpha = math.min (math.max (0, (total_stop_time - reel_stop_time) / impl.blur_show_time), 1)
      alpha        = 1 - blured_alpha

      if reel_stop_time < total_stop_time then
        reel_position = reel_stop_time * speed
      else
        if not reel.stopping_started then
          if impl.stop_sounds and impl.stop_sounds [reel_index] then
            PlaySound (impl.stop_sounds [reel_index])
          else
            PlaySound (impl.stop_sound)
          end

          reel.stopping_started = true
        end
      
        local finished = true
      
        if impl.stop_spline then
          local spline_time = reel_stop_time - total_stop_time
        
          if impl.stop_spline.MaxTime > spline_time then
            finished = false
            
            reel_position = -reel.stop_length * spline_factor + impl.line_interval * impl.stop_spline:Eval (spline_time) * spline_factor
          end
        end
        
        if finished then
          reel.spin_stop_start_time = nil
          reel.spinning             = false
          
          stop_node:Unbind ()
          stop_node:UnbindAllChildren ()
          node:BindToParent (game.current_screen.node)
          node.Position = reel.start_position + vec3 (0, impl.line_interval * (reel.stop_position - 1), 0)
          
          alpha        = 1
          blured_alpha = 0
          
          if reel.items_alpha ~= alpha then
            reel.items_alpha = alpha

            for i = 1, #reel.items do
              local item = reel.items [i]
    
              item.sprite.Alpha        = alpha
              item.blured_sprite.Alpha = blured_alpha
            end
          end
          
          reel.stop_items = nil
    
          UpdateReelGroups (impl, reel)
          HideInvisibleItems (impl, reel, nil, alpha)
          
          if impl.event_handler and impl.event_handler.OnReelSpinStopped then
            impl.event_handler.OnReelSpinStopped (reel_index)
          end
                
          return
        end
      end
    else
      reel_position = spinup_distance + full_speed_reel_time * speed 
    end
  end

  if reel.spin_stop_start_time then
    stop_node.Position = reel.stop_start_position + vec3 (0, reel_position, 0)

    if reel.stop_items_alpha ~= alpha then
      reel.stop_items_alpha = alpha

      for i = 1, #reel.stop_items do
        local item = reel.stop_items [i]
    
        item.sprite.Alpha        = alpha
        item.blured_sprite.Alpha = blured_alpha
      end
    end

    if impl.visible_lines_count then
      local top_visible_item = #reel.stop_items + math.floor ((reel_position + reel.stop_start_offset) / impl.line_interval) - 2
      local hide_item        = top_visible_item + impl.visible_lines_count

      if hide_item > 0 and hide_item <= #reel.stop_items then
        reel.stop_items [hide_item].sprite.Alpha        = 0
        reel.stop_items [hide_item].blured_sprite.Alpha = 0
      end
    end
  else
    node.Position = reel.start_position + vec3 (0, ClampReelPosition (reel, reel_position + reel.spin_start_position), 0)

    if reel.items_alpha ~= alpha then
      reel.items_alpha = alpha

      for i = 1, #reel.items do
        local item = reel.items [i]
    
        item.sprite.Alpha        = alpha
        item.blured_sprite.Alpha = blured_alpha
      end
    end
    
    UpdateReelGroups (impl, reel)
    HideInvisibleItems (impl, reel, 1, alpha)
  end
end

local function OnSpinFinished (impl)
  impl.spin_started = false
  impl.needs_stop_spinning = false
  
  StopStoredSound (impl.spin_sound_name)
  
  if impl.on_stop_handler then
    impl.on_stop_handler ()
  end
end

local function CancelSpin (impl)
  impl.spin_started = false
  impl.needs_stop_spinning = false

  if impl.spin_sound_name then
    StopStoredSound (impl.spin_sound_name)
  end
end

local function Update (impl)
  if not impl.spin_started then return end

  local spin_time = game.current_screen.timers.default:GetTime () - impl.spin_start_time

  local first_unlocked_reel = 1
  local last_unlocked_reel  = GetLastUnlockedReel (impl)
  
  for i = 1, #impl.reels do
    if impl.spinning_directions [i] ~= nil then
      first_unlocked_reel = i
      break
    end
  end

  if impl.needs_stop_spinning and spin_time > impl.spin_min_duration and not impl.spin_stopping then
    impl.spin_stopping = true
    impl.reels [first_unlocked_reel].spin_stop_start_time = game.current_screen.timers.default:GetTime ()

    if impl.event_handler and impl.event_handler.OnReelStopInitiated then
      impl.event_handler.OnReelStopInitiated (first_unlocked_reel)
    end
  
    PrepareReelForStopping (impl, first_unlocked_reel)
  end

  for i = 1, #impl.reels do
    if impl.spinning_directions [i] ~= nil then
      UpdateReelSpin (impl, i, spin_time)
    end
  end    
    
  if not impl.reels [last_unlocked_reel].spinning then
    OnSpinFinished (impl)
  end
end

local function GetReelOffsetForPosition (impl, position)
  return position * impl.line_interval
end

local function ScrollReel (impl, reel, position)
  local reel      = impl.reels [reel]
  local reel_node = reel.node

  reel_node.Position = reel.start_position + vec3 (0, GetReelOffsetForPosition (impl, position), 0)
    
  UpdateReelGroups (impl, reel)
  
  HideInvisibleItems (impl, reel, nil, 1)
end

local function ActivateBonusTurboMode (impl)
  if not impl.spin_started then return end
  if impl.turbo_mode_activated then return end
  
  impl.turbo_mode_activated = true
  
  if impl.turbo_sound then
    StopStoredSound (impl.spin_sound_name)
    
    impl.spin_sound_name = impl.turbo_sound

    PlayStoredSound (impl.spin_sound_name)
  end
  
  impl.spin_speed = impl.spin_speed * 2
end

local function PrepareForBonusTurboMode (impl)
  if not impl.spin_started then return end
  if impl.turbo_mode_activated then return end
  
  impl.spin_stop_delay = BONUS_TURBO_SPIN_STOP_DELAY
end

local function BindReels (impl)
  for i = 1, #impl.reels do
    impl.reels [i].node:BindToParent (game.current_screen.node)
  end
end

local function UnbindReels (impl)
  for i = 1, #impl.reels do
    impl.reels [i].node:Unbind ()
  end
end

local function SetSpinningDirections (impl, directions) 
  impl.spinning_directions = directions
end

local function Setup (impl, reels, line_interval, xml_desc, visible_lines_count)
  impl.reels               = reels
  impl.line_interval       = line_interval
  impl.visible_lines_count = visible_lines_count
  impl.spinning_directions = { true, true, true, true, true }

  impl.stop_sound              = xml_desc:Get ("StopSound")
  impl.blur_show_time          = xml_desc:Get ("BlurShowTime")
  impl.normal_spin_start_delay = xml_desc:Get ("SpinStartDelay") / 1000
  impl.normal_spin_speed       = -xml_desc:Get ("SpinSpeed") * impl.line_interval
  impl.normal_spin_stop_delay  = xml_desc:Get ("SpinStopDelay") / 1000
  impl.normal_spin_delay       = xml_desc:Get ("SpinDelay") / 1000
  impl.fast_spin_start_delay   = xml_desc:Get ("FastSpinStartDelay", tostring (xml_desc:Get ("SpinStartDelay") / 2)) / 1000
  impl.fast_spin_speed         = -xml_desc:Get ("FastSpinSpeed", tostring (xml_desc:Get ("SpinSpeed") * 2)) * impl.line_interval
  impl.fast_spin_stop_delay    = xml_desc:Get ("FastSpinStopDelay", tostring (xml_desc:Get ("SpinStopDelay") / 2)) / 1000
  impl.fast_spin_delay         = xml_desc:Get ("FastSpinDelay", tostring (xml_desc:Get ("SpinDelay") / 4)) / 1000

  local turbo_sound = xml_desc:Get ("TurboSound", "")

  if turbo_sound ~= "" then
    impl.turbo_sound = turbo_sound
  end

  local spin_start_anim = xml_desc:Get ("SpinStartAnim")
  
  if spin_start_anim == "instantly" then
    impl.reel_spinup_time                = 0
    impl.normal_reel_spinup_acceleration = 9999
    impl.fast_reel_spinup_acceleration   = 9999
  elseif spin_start_anim == "default" then
    impl.reel_spinup_time                = DEFAULT_SPINUP_TIME
    impl.normal_reel_spinup_acceleration = impl.normal_spin_speed / impl.reel_spinup_time
    impl.fast_reel_spinup_acceleration   = impl.fast_spin_speed / impl.reel_spinup_time
  else
    error ("Unsupported start animation type " .. spin_start_anim)
  end

  local spin_stop_anim = xml_desc:Get ("SpinStopAnim")
  
  if spin_stop_anim == "instantly" then
  elseif spin_stop_anim == "spline" then
    impl.stop_spline = Math.SplineLoader.LoadLinearFloatSpline (xml_desc:Find ("StopSpline"), "Key", "Time", "Value")
  else
    error ("Unsupported stop animation type " .. spin_stop_anim)
  end

  impl.normal_spinup_time          = impl.normal_spin_start_delay * (#impl.reels - 1) + impl.reel_spinup_time
  impl.normal_spin_min_duration    = impl.normal_spin_delay + impl.normal_spinup_time
  impl.normal_reel_spinup_distance = impl.normal_reel_spinup_acceleration * impl.reel_spinup_time * impl.reel_spinup_time / 2
  impl.fast_spinup_time            = impl.fast_spin_start_delay * (#impl.reels - 1) + impl.reel_spinup_time
  impl.fast_spin_min_duration      = impl.fast_spin_delay + impl.fast_spinup_time
  impl.fast_reel_spinup_distance   = impl.fast_reel_spinup_acceleration * impl.reel_spinup_time * impl.reel_spinup_time / 2
end

function CreateSpinEngine (on_stop_handler, event_handler)
  local impl = {}  

  OnSpinFinished (impl)
  
  impl.on_stop_handler = on_stop_handler
  impl.event_handler   = event_handler

  local public_methods = {}

  public_methods.ActivateBonusTurboMode   = function () ActivateBonusTurboMode (impl) end
  public_methods.BindReels                = function () BindReels (impl) end
  public_methods.CancelSpin               = function () CancelSpin (impl) end
  public_methods.PrepareForBonusTurboMode = function () PrepareForBonusTurboMode (impl) end
  public_methods.ReelsCount               = function () return #impl.reels end
  public_methods.ScrollReel               = function (reel, position) ScrollReel (impl, reel, position) end
  public_methods.SetReelsScissor          = function (scissor) SetReelsScissor (impl, scissor) end
  public_methods.SetSpinningDirections    = function (directions) SetSpinningDirections (impl, directions) end
  public_methods.Setup                    = function (reels, line_interval, xml_desc, visible_lines_count) Setup (impl, reels, line_interval, xml_desc, visible_lines_count) end
  public_methods.StartSpin                = function (spin_sound_name, fast_mode) StartSpin (impl, spin_sound_name, fast_mode) end
  public_methods.StopSpin                 = function (stop_positions, stop_sounds) StopSpin (impl, stop_positions, stop_sounds) end
  public_methods.UnbindReels              = function () UnbindReels (impl) end
  public_methods.Update                   = function () Update (impl) end

  return public_methods
end
