local TRANSACTIONS_DB_PATH = APP_DATA_PATH .. "pending_transactions.xml"

local function FindTransactionDesc (impl, transaction)
  if not Common.File.IsFileExist (TRANSACTIONS_DB_PATH) then return end

  local db                = Common.StringNode.LoadXml (TRANSACTIONS_DB_PATH)
  local db_children_count = db.ChildrenCount
  
  for i = db_children_count - 1, 0, -1 do
    local transaction_node = db:Child (i)
    
    if transaction_node.Name == "Transaction" then
      local product = transaction_node:Get ("Product")
      
      if product == transaction.ProductId then
        return { login_id = transaction_node:Get ("LoginId"), network = transaction_node:Get ("Network") }
      end
    end
  end
end

local function RemoveTransactionFromDb (impl, transaction)
  impl.pending_transactions [transaction.Id] = nil

  if not Common.File.IsFileExist (TRANSACTIONS_DB_PATH) then return end

  local db                = Common.StringNode.LoadXml (TRANSACTIONS_DB_PATH)
  local db_children_count = db.ChildrenCount
  
  for i = db_children_count - 1, 0, -1 do
    local transaction_node = db:Child (i)
    
    if transaction_node.Name == "Transaction" then
      local product  = transaction_node:Get ("Product")
      
      if product == transaction.ProductId then
        db:RemoveChild (i)
        break
      end
    end
  end
  
  db:SafeSaveXml (TRANSACTIONS_DB_PATH)
end

local function OnPaymentFailed (impl, transaction)
  if impl.waiting_for_product_transaction and impl.waiting_for_product_transaction.Id == transaction.Id then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("PURCHASING_PRODUCT_FAILED_LINE_1") }, nil, nil, ApplicationDataGet (OS_TYPE) ~= "Android")
    
    impl.waiting_for_product_transaction = nil
  end
end

local function OnTransactionFinished (impl, transaction, purchase_status, state, error)
  if not state or not purchase_status then
    OnPaymentFailed (impl, transaction)
  else
    if impl.waiting_for_product_transaction and impl.waiting_for_product_transaction.Id == transaction.Id then
      game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("PURCHASING_PRODUCT_SUCCEEDED_LINE_1") }, nil, nil, true)
    
      impl.waiting_for_product_transaction = nil
    
      local flurry_buy_event_data = ApplicationDataGet (FLURRY_BUY_EVENT_DATA)

      if flurry_buy_event_data then
        LogFlurryEvent ("BankBuySucceeded" .. flurry_buy_event_data.tab_name .. flurry_buy_event_data.package_index, {}, true)
        LogFlurryEvent ("BankBuySucceeded" .. flurry_buy_event_data.tab_name, { PackageIndex = flurry_buy_event_data.package_index }, true)
      end
    end  
  end
    
  RemoveTransactionFromDb (impl, transaction)
end

local function OnTransactionUpdated (impl, transaction)
  impl.pending_transactions [transaction.Id] = transaction

  if not impl.payment_url then return end

  if transaction.State == Store.TransactionState.Failed then
    print ("Finishing failed transaction (reason '" .. transaction.Status .. "')")

    transaction:Finish (Store.Transaction.CreateOnFinishedCallback (function (state, error)
      OnTransactionFinished (impl, transaction, false, state, error)
    end))
    
    local flurry_buy_event_data = ApplicationDataGet (FLURRY_BUY_EVENT_DATA)
    
    if flurry_buy_event_data then
      LogFlurryEvent ("BankBuyFailed" .. flurry_buy_event_data.tab_name .. flurry_buy_event_data.package_index, {}, true)
      LogFlurryEvent ("BankBuyFailed" .. flurry_buy_event_data.tab_name, { PackageIndex = flurry_buy_event_data.package_index }, true)
    end
  elseif transaction.State == Store.TransactionState.Purchased or transaction.State == Store.TransactionState.Restored then
    local transaction_desc = FindTransactionDesc (impl, transaction)
  
    if not transaction_desc then
      OnPaymentFailed (impl, transaction)
      return
    end
  
    local message_dictionary = { loginId = transaction_desc.login_id, network = transaction_desc.network, transactionReceipt = transaction.Receipt }

    if ApplicationDataGet (OS_TYPE) == "Android" and impl.store_name ~= AMAZON_STORE_NAME then
      local is_phone = 0
    
      if GetPlatform () == "phone" then
        is_phone = 1
      end
      
      message_dictionary.phone = is_phone
    end
    
    local transaction_properties = transaction.Properties

    if transaction_properties:IsPresent ("Signature") then
      message_dictionary.signature = transaction_properties:GetString ("Signature")
    end

    if impl.store_name == AMAZON_STORE_NAME then
      if transaction_properties:IsPresent ("UserId") then
        message_dictionary.userId = transaction_properties:GetString ("UserId")
      end
    end

    local message = JsonCoder.encode (message_dictionary)

    if DEBUG then
      print ("Checking store response " .. message)
    end

    Common.File.AsyncPostString (impl.payment_url, message, Common.File.CreateAsyncLoadHandler (function (answer_string)
      if DEBUG then
        print ("payment check response = '" .. answer_string .. "'")
      end

      local node         = Common.StringNode.LoadJsonFromString (answer_string)
      local events_count = node.ChildrenCount
  
      local success               = false
      local transaction_processed = false
  
      for i = 0, events_count - 1 do
        local event = node:Child (i)
        local event_type = event:Find ("event")
    
        if event_type then
          local event_type = event_type:Attribute (0)
      
          if event_type == "login.payment" then
            local status = event:Find ("status"):Attribute (0)
            
            if status == "success" then
              success               = true
              transaction_processed = true
            elseif status == "error" then
              local error_id = event:Find ("errorId"):Attribute (0)
              
              if error_id == "10" then
                transaction_processed = true
              end
            end

            break
          end
        end
      end

      if transaction_processed then
        transaction:Finish (Store.Transaction.CreateOnFinishedCallback (function (state, error)
          OnTransactionFinished (impl, transaction, success, state, error)
        end))
      end
      
      if success then
        HsConnectionSendMessage (1, { event = "payment.check" })
      else
        print ("payment check failed")
        OnPaymentFailed (impl, transaction)
      end
    end))
  end
end

local function BuyProduct (impl, id)
  if not impl.store then return end
  if not impl.initialized then return end

  local db
  
  if Common.File.IsFileExist (TRANSACTIONS_DB_PATH) then
    db = Common.StringNode.LoadXml (TRANSACTIONS_DB_PATH)
  else
    db = Common.StringNode.Create ()
    
    db.Name = "PendingTransactions"
  end 
  
  local transaction = Common.StringNode.Create ()
        
  transaction.Name = "Transaction"
        
  transaction:Set ("Product", tostring (id))
  transaction:Set ("LoginId", tostring (GetSocialManager ().GetUserId ()))
  transaction:Set ("Network", tostring (GetSocialManager ().GetNetworkName ()))
        
  db:AddChild (transaction)

  db:SafeSaveXml (TRANSACTIONS_DB_PATH)  

  for _, value in pairs (impl.pending_transactions) do
    if value.ProductId == id then
      impl.waiting_for_product_transaction = value
      OnTransactionUpdated (impl, value)
      return
    end
  end
  
  local buy_properties = Common.PropertyMap.Create () 
  
  buy_properties:SetInteger ("Consumable", 1)
  
  impl.waiting_for_product_transaction = impl.store:BuyProduct (id, 1, buy_properties)
end

local function CanBuyProducts (impl)
  if not impl.store then return true end
  if not impl.initialized then return false end
  
  return impl.store.CanBuyProducts
end

local function OnPaymentUrlUpdated (impl, url)
  impl.payment_url = url

  local transactions = {}
  
  for _, value in pairs (impl.pending_transactions) do
    table.insert (transactions, value)
  end
  
  for i = 1, #transactions do
    OnTransactionUpdated (impl, transactions [i])
  end
end

local function GetStoreName (impl)
  if not impl.initialized then return nil end
  
  return impl.store_name 
end

local function OnStoreInitialized (impl, store, store_name)
  impl.initialized = true

  impl.store      = store
  impl.store_name = store_name
    
  if impl.transaction_update_connection then
    impl.transaction_update_connection:Disconnect ()
  end

  impl.transaction_update_connection = impl.store:RegisterTransactionUpdateHandler (Store.Store.CreatePurchaseCallback (function (transaction)
    OnTransactionUpdated (impl, transaction)
  end))
end

local function CreateStore (impl, stores_names, index)
  if index > #stores_names then return end

  local store_name = stores_names [index]

  Store.Store.Create (store_name, Store.Store.CreateOnInitializedCallback (function (store)
    if not store.CanBuyProducts then
      if index < #stores_names then
        CreateStore (impl, stores_names, index + 1)
      else
        OnStoreInitialized (impl, store, store_name)
      end

      return
    end
  
    OnStoreInitialized (impl, store, store_name)
  end))
end

local function CreateStoreManager ()
  local impl = {}
  
  impl.initialized = false
  
  impl.pending_transactions = {}

  local stores_names = {}

  local os = ApplicationDataGet (OS_TYPE)

  if string.find (os, "iOS") ~= nil then
    table.insert (stores_names, "IOSStore")
  elseif os == "Android" then
    local system_properties = System.Application.SystemProperties

    table.insert (stores_names, "AndroidStore")

    if system_properties:GetString ("Manufacturer") == "Amazon" then --amazon store initialized after google play store in case of google play devices with 'Amazon' manufacturer
      table.insert (stores_names, AMAZON_STORE_NAME)
    end
  end

  CreateStore (impl, stores_names, 1)

  local public_methods = {}

  public_methods.BuyProduct          = function (id) BuyProduct (impl, id) end
  public_methods.CanBuyProducts      = function () return CanBuyProducts (impl) end
  public_methods.GetStoreName        = function () return GetStoreName (impl) end 
  public_methods.OnPaymentUrlUpdated = function (url) OnPaymentUrlUpdated (impl, url) end
    
  return public_methods
end

local StoreManagerInstance

function GetStoreManager ()
  if not StoreManagerInstance then
    StoreManagerInstance = CreateStoreManager ()
  end
  
  return StoreManagerInstance
end
