local DEFAULT_CALL_DELAY = 1 / 60 --   ,     

--  
local function StartPeriodicCallImpl (impl, delay, func, args)
  if not args then args = {} end

  local action = {}  
  
  local timer   = Common.Timer.Create (impl.timer)
  local active  = true
  local stopped = false
  
  timer.AutoUpdate = false
  timer.Started    = true

  action.IsStopped = function ()
    return stopped
  end

  action.Stop = function ()
    if action.OnStop then
      action.OnStop ()
    end
    
    active  = false
    stopped = true
  end

  action.Pause = function()
    if stopped then return end
   
    timer:Stop ()
    active = false
  end
  
  action.Resume = function()
    if stopped then return end
   
    timer:Start ()
    active = true
  end
  
  action.GetTime = function ()
    return timer.Time
  end

  local trigger_count = 0

  action.Update = function ()
    if not active then return end
   
    timer:Update ()    
    
    local count = math.floor (timer.Time / delay)    
    
    if count <= trigger_count then return end

    trigger_count = count    
    
    if func (unpack (args)) == false then
      action.Stop ()
    end
  end

  table.insert (impl.actions, action)

  return action
end

--  
local function SlowActionImpl (impl, action_time, func, delay)
  
  if not action_time then return end

  if not delay then delay = DEFAULT_CALL_DELAY end

  local action
  
  action = StartPeriodicCallImpl (impl, delay, function()
    local ellapsed_time = action.GetTime ()
    local status        = ellapsed_time / action_time

    if status < 1 then 
      func (status)
    else
      func (1)
      action.Stop ()
    end
  end)

  return action
end

--  
local function DelayedActionImpl (impl, delay, func, args) 
  local action
  
  local function Update ()
    action.Stop ()
    func ()
  end

  action = StartPeriodicCallImpl (impl, delay, Update, args)

  return action
end

--  
local function Update (impl)
  impl.timer:Update ()

  local current_actions = impl.actions
  
  impl.actions = {}
  
  for _, action in pairs (current_actions) do
    action.Update ()
    
    if not action.IsStopped () then
      table.insert (impl.actions, action)
    end
  end
end

--  
local function ResetTimer (impl)
  impl.actions = {}

  impl.timer:Reset ()  
end

--  
function CreateTimer (auto_update, started, parent_timer)
  local impl = {}

  if parent_timer then
    impl.timer = parent_timer.CreateChildTimer ()
  else
    impl.timer = Common.Timer.Create ()
  end

  impl.timer.AutoUpdate = auto_update
  impl.timer.Started    = started

  impl.actions = {}

  local methods = {}

  methods.StartPeriodicCall = function (delay, func, ...) return StartPeriodicCallImpl (impl, delay, func, { ... }) end
  methods.SlowAction        = function (action_time, func, delay) return SlowActionImpl (impl, action_time, func, delay) end
  methods.DelayedAction     = function (delay, func, ...) return DelayedActionImpl (impl, delay, func, { ... }) end
  methods.Update            = function () Update (impl) end
  methods.CreateChildTimer  = function () return Common.Timer.Create (impl.timer) end
  methods.Stop              = function () impl.timer:Stop () end
  methods.Start             = function () impl.timer:Start () end
  methods.Reset             = function () ResetTimer (impl) end
  methods.IsStarted         = function () return impl.timer.Started end
  methods.GetTime           = function () return impl.timer.Time end  

  return methods
end

--    

function StartPeriodicCall (timer_name, delay, func, ...)
  local timer = game.current_screen.timers [timer_name]
  
  if not timer then
    error (string.format ("Timer '%s' not found", timer_name))
  end  

  return timer.StartPeriodicCall (delay, func, ...)
end

--  
function SlowAction (timer_name, action_time, func, delay)
  local timer = game.current_screen.timers [timer_name]
    
  if not timer then
    error (string.format ("Timer '%s' not found", timer_name))
  end

  return timer.SlowAction (action_time, func, delay)
end

--  
function DelayedAction (timer_name, delay, func, ...) 
  local timer = timer_name
  
  if type (timer_name) == "string" then
    timer = game.current_screen.timers [timer_name]
  end
  
  if not timer then
    error (string.format ("Timer '%s' not found", timer_name))
  end

  return timer.DelayedAction (delay, func, ...)
end
