URL_SPRITE_MODE_CACHE_FILE = 0

local EPS                     = 0.0001
local SUPPORTED_IMAGE_FORMATS = { "png", "jpg", "bmp", "tga", "tif" }
local URL_SPRITES_CACHE_SIZE  = 20
local TEXTURE_DESC_TEMPLATE   = "<?xml version='1.0' encoding='utf-8'?> <texture_desc size_x='%f' size_y='%f'/>"
local XMTL_TEMPLATE           = "<?xml version='1.0' encoding='utf-8'?> <material_library> <materials> <material id='%s' program='flat' tags='transparent'> <texmap semantic='image' sampler='default' image='%s'/> </material> </materials> </material_library>"

local url_sprite_tmp_file_index = 0

local url_sprites_cache = {}

local function GetNextHigherPowerOfTwo (value)
  local power = math.ceil (math.log10 (value) / math.log10 (2))
  
  return math.pow (2, power)
end

local function PerformHide (impl, dont_unload)
  if impl.shown == false then return end

  for i = 1, #url_sprites_cache do
    if url_sprites_cache [i] == impl then
      table.remove (url_sprites_cache, i)
      break
    end 
  end
  
  if impl.download_action then
    impl.download_action.Stop ()
    impl.download_action = nil
  end

  impl.shown = false
  
  impl.activity_indicator.Stop ()
  impl.activity_indicator.GetSprite ():Unbind ()

  impl.resource_binding = nil

  if impl.sprite.Parent then
    impl.sprite.Material = "transparent"

    if not dont_unload then
      local position = impl.sprite.Position

      impl.sprite.WorldPosition = vec3 (100, -100, impl.sprite.WorldPosition.z)

      ExecuteOnNextUpdate (function ()
        ExecuteOnNextUpdate (function ()
          UnloadUnusedResources ()
        end)
      
        impl.sprite.Position = position
      end)
    end
  else
    UnloadUnusedResources ()
  end
  
  if impl.download_path and impl.downloading then
    impl.downloading = nil
  
    GetDownloadsManager ().CancelDownloadFile (impl.download_path)
  end
end

local function OnHide (impl, force_unload)
  if impl.shown == false then return end
  
  if force_unload then
    PerformHide (impl)
    return
  end

  for i = 1, #url_sprites_cache do
    if url_sprites_cache [i] == impl then return end   --already in cache 
  end
  
  table.insert (url_sprites_cache, impl)
  
  if #url_sprites_cache > URL_SPRITES_CACHE_SIZE then
    PerformHide (url_sprites_cache [1])
  end
end

local function OnShow (impl)
  for i = 1, #url_sprites_cache do
    if url_sprites_cache [i] == impl then
      table.remove (url_sprites_cache, i)
      break
    end 
  end
  
  if impl.shown then return end

  impl.shown = true

  impl.sprite.Material  = impl.default_material.material
  impl.sprite.TexSize   = impl.default_material.tex_size
  impl.sprite.TexOffset = impl.default_material.tex_offset

  if not impl.material then return end
  
  impl.activity_indicator.Start ()
  impl.activity_indicator.GetSprite ():BindToParent (impl.sprite, Scene.NodeBindMode.Default, Scene.NodeTransformSpace.Local)

  impl.download_action = DelayedAction ("ui", 0.3, function ()
    impl.download_action = nil
  
    impl.downloading = true
  
    GetDownloadsManager ().DownloadFile (impl.url, impl.download_path, function (status, arg2)
      if status == Common.BackgroundCopyStateStatus.Failed then 
        impl.downloading = nil
        return 
      end
    
      if status ~= Common.BackgroundCopyStateStatus.Finished then return end

      impl.downloading = nil

      local texture_size

      if not Common.File.IsFileExist (impl.material_file_path) then
        local image        = Media.Image.Create (impl.download_path)
        local image_width  = image.Width
        local image_height = image.Height
        local image_ar     = image_width / image_height
        local sprite_ar    = impl.size.x / impl.size.y

        local resized_image_width
        local resized_image_height
        local source_area_width
        local source_area_height

        if sprite_ar > image_ar then
          resized_image_width  = math.min (image_width, impl.size.x)
          resized_image_height = resized_image_width / image_ar
          source_area_width    = resized_image_width
          source_area_height   = source_area_width / sprite_ar
        else
          resized_image_height = math.min (image_height, impl.size.y)
          resized_image_width  = resized_image_height * image_ar
          source_area_height   = resized_image_height
          source_area_width    = source_area_height * sprite_ar
        end
        
        resized_image_width  = math.ceil (math.max (1, resized_image_width))
        resized_image_height = math.ceil (math.max (1, resized_image_height))
        source_area_width    = math.ceil (math.max (1, source_area_width))
        source_area_height   = math.ceil (math.max (1, source_area_height))
      
        image:Resize (resized_image_width, resized_image_height, 1)
    
        local target_image_width  = GetNextHigherPowerOfTwo (source_area_width)
        local target_image_height = GetNextHigherPowerOfTwo (source_area_height)

        local target_image = Media.Image.Create (target_image_width, target_image_height, 1, Media.ImageFormat.RGBA8)
     
        target_image:FillRGBA (vec4 (0, 0, 0, 0))
    
        local put_from_x = 0
        local put_from_y = 0

        if sprite_ar > image_ar then
          put_from_y = math.floor ((resized_image_height - source_area_height) / 2)
        else
          put_from_x = math.floor ((resized_image_width - source_area_width) / 2)
        end
        
        target_image:PutImage (0, 0, 0, source_area_width, source_area_height, 1, image, put_from_x, put_from_y, 0)
    
        texture_size = vec2 (source_area_width / target_image_width, source_area_height / target_image_height)
    
        local save_path = impl.image_file_path
    
        target_image:Save (save_path)

        image        = nil
        target_image = nil
        collectgarbage ("collect")
    
        local xml_string = string.format (TEXTURE_DESC_TEMPLATE, texture_size.x, texture_size.y)
        
        Common.File.PostString (impl.texture_desc_file_path, xml_string)
    
        local xmtl_string = string.format (XMTL_TEMPLATE, impl.material, save_path)
        
        Common.File.PostString (impl.material_file_path, xmtl_string)
      else
        local xml = Common.StringNode.LoadXml (impl.texture_desc_file_path)
  
        texture_size = vec2 (xml:Get ("size_x") + 0, xml:Get ("size_y") + 0)
      end
      
      impl.resource_binding = LoadResources ({ impl.image_file_path, impl.material_file_path })
      impl.sprite.Material  = impl.material
      impl.sprite.TexOffset = vec2 (0, 0)
      impl.sprite.TexSize   = texture_size
    
      impl.activity_indicator.Stop ()
      impl.activity_indicator.GetSprite ():Unbind ()
    end, false)
  end)
end

local function LoadImage (impl, url, show)
  if impl.url and impl.url == url then return end

  if show then
    PerformHide (impl, true)
  end

  local path_extension = GetPathExtension (url)

  local supported = false

  for i = 1, #SUPPORTED_IMAGE_FORMATS do
      if path_extension == SUPPORTED_IMAGE_FORMATS [i] then
        supported = true
        break
      end
    end

  if supported then
    local tmp_file_name = CACHE_PATH .. UrlToFileName (url)

    impl.download_path          = tmp_file_name .. "." .. path_extension
    impl.image_file_path        = tmp_file_name .. impl.size.x .. impl.size.y .. "." .. path_extension .. ".tga"
    impl.material_file_path     = tmp_file_name .. impl.size.x .. impl.size.y .. ".xmtl"
    impl.texture_desc_file_path = tmp_file_name .. impl.size.x .. impl.size.y .. ".xml"
    impl.material               = GetLastPathComponent (tmp_file_name) .. impl.size.x .. impl.size.y

    impl.url  = url
  else
    impl.image_file_path        = nil
    impl.material_file_path     = nil
    impl.texture_desc_file_path = nil
    impl.material               = nil
  end
  
  if show then
    OnShow (impl)
  end
end

function CreateUrlSprite (mode, url, params, default_material, size, activity_indicator_size, screen_scale)
  local impl = {}

  screen_scale = screen_scale or 1

  impl.size = { x = math.ceil (size.x * screen_scale), y = math.ceil (size.y * screen_scale) }
--  impl.size = { x = 32, y = 33 }

  LoadImage (impl, url, false)  

  impl.mode = mode

  impl.default_material = default_material or { material = "transparent", tex_size = vec2 (1, 1), tex_offset = vec2 (0, 0) }

  impl.sprite = Scene.Sprite.Create ()

  impl.sprite.Material = "transparent"
  impl.sprite.Scale    = vec3 (size.x, size.y, 1)

  impl.activity_indicator = CreateActivityIndicator (activity_indicator_size, "ui")

  impl.activity_indicator.GetSprite ().Position = vec3 (0, 0, -EPS)

  PerformHide (impl)

  local public_methods = {}

  public_methods.GetActivityIndicator = function () return impl.activity_indicator end
  public_methods.GetSprite            = function () return impl.sprite end
  public_methods.GetUrl               = function () return impl.url end
  public_methods.LoadImage            = function (url) LoadImage (impl, url, true) end
  public_methods.OnHide               = function (force_unload) OnHide (impl, force_unload) end
  public_methods.OnShow               = function () OnShow (impl) end
    
  return public_methods
end

function ClearUrlSpritesCache ()
  for i = #url_sprites_cache, 1, -1 do
    OnHide (url_sprites_cache [i], true)
  end
  
  url_sprites_cache = {}
end
